#ifndef FKTDEBUG
#define FKTDEBUG 0
#endif

//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : FKTPARS.Y
//
// Description:
//		Parser of function compiler.
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <ccpptyp.h>
#include <stdlib.h>
#include "defines.h"
#include "ownerr.h"
#include "Node.h"
#include "SimplVar.h"
#include "ArrayVar.h"
#include "HashTab.h"
#include "FKTScan.h"
#include "FKTPars.h"

// global variables declared in FUNCTION.CPP
extern Node *_compiled_function;
extern HashTable FKT_LongVariables;

// prototypes
INT WordRep(CHAR);
void FKT_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno);
void FKT_declare_arrayvariable(char *varid,unsigned int size,unsigned int lineno);
char FKT_test_type_variable(char *name,char type,unsigned int lineno);

extern FKTSTYPE fktlval;
INT WordRep(CHAR ch)
{
	return(ch-(islower(ch) ? 'a' : 'A'));
}

void FKT_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno)
{
	SimpleVariable *help;

	if (FKT_LongVariables.Search(varid,VAR_SIMPLE)!=NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"FKTPARS: variable %s already declared !",varid);
		warning(msg,lineno);
		dealloc_msgmem(msg);
	}
	else
	{
		if ((help=new SimpleVariable(varid,initvalue))==NULL)
			fatal("FKTPARS: out of memory!!",lineno);
		FKT_LongVariables.Insert(help);
	}
}

void FKT_declare_arrayvariable(char *varid,unsigned int size,unsigned int lineno)
{
	ArrayVariable *help;

	if (FKT_LongVariables.Search(varid,VAR_ARRAY)!=NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"FKTPARS: variable %s already declared !",varid);
		warning(msg,lineno);
		dealloc_msgmem(msg);
	}
	else
	{
		if ((help=new ArrayVariable(varid,size))==NULL)
			fatal("FKTPARS: out of memory!!",lineno);
		FKT_LongVariables.Insert(help);
	}
}

char FKT_test_type_variable(char *name,char type,unsigned int lineno)
{
	ListElement *help;
	if (FKT_LongVariables.Search(name,type)==NULL)
	{
		char *msg;
		if ((msg=new char[ERR_MSG_LEN])==NULL)
			fatal("FKTPARS: out of memory!!");
		sprintf(msg,"FKTPARS: variable %s not declared or type mismatch!!",name);
		FKT_error(msg,lineno);
		delete msg;
		return(FALSE);
	}
	return(TRUE);
}
static short fktdef[] = {
	  -1,    9,    8,    5,    3,    7,    6,   14,   13,   12,
	  11,   10,    4
};
static short fktex[] = {
	   0,    0,   -1,    1
};
static short fktact[] = {
	 -56,  -55,  258,  256,  -73,   59,  -46,  -44,  -47,  -49,
	 -50,  -53,  -54,  -48,  -51,  -52,  273,  272,  268,  265,
	 264,  262,  261,  260,  259,  123,  -42,  -62,  -63,  -61,
	 -60,  -41,  279,  273,  272,  271,  270,   40,  -40,   61,
	 -39,   61,  -38,   40,  -46,  -47,  -49,  -50,  -53,  -54,
	 -48,  -51,  -52,  273,  272,  268,  265,  264,  262,  261,
	 260,  123,  -36,   40,  -81,   59,   -5,  273,  -32,   40,
	 -27,  -28,  -29,  -30,  -31,  302,  277,  276,  275,  274,
	 -25,  260,  -46,  -68,  -47,  -49,  -50,  -53,  -54,  -48,
	 -51,  -52,  273,  272,  268,  265,  264,  262,  261,  260,
	 125,  123,  -22,  -23,   91,   61,  -21,  -78,   59,   44,
	 -58,  -27,  -28,  -29,  -30,  -31,  302,  277,  276,  275,
	 274,   41,  -19,  -27,  -28,  -29,  -30,  -31,  302,  277,
	 276,  275,  274,   41,  -18,   40,  -17,  -27,  -28,  -29,
	 -30,  -31,  302,  277,  276,  275,  274,   41,  -69,  -70,
	 271,  270,  -59,  -27,  -28,  -29,  -30,  -31,  302,  277,
	 276,  275,  274,   41,  -27,  -28,  -29,  -30,  277,  276,
	 275,  274,  -30,  277,  -29,  -30,  277,  276,  -28,  -29,
	 -30,  277,  276,  275,  -71,   93,  -14,  263,  -66,  -27,
	 -28,  -29,  -30,  -31,  302,  277,  276,  275,  274,   41,   -1
};
static short fktpact[] = {
	  26,   75,   75,   75,  104,   75,   75,  168,  173,  173,
	 176,  181,  187,   53,  194,  185,   53,   32,   53,  158,
	  67,  150,  150,  142,  135,  128,   32,   32,   32,   32,
	  32,   32,  116,  108,   92,   32,   81,   32,   32,   32,
	  69,   32,   16,   67,   65,   53,   63,   53,   43,   32,
	  41,   39,   32,   32,   16,    5,    2
};
static short fktgo[] = {
	 -15,  -12,  -11,  -10,   -9,   -8,  -20,  -24,  -26,   -7,
	  -6,  -33,   -4,   -3,   -2,   52,   49,   41,   39,   38,
	  37,   35,   31,   30,   29,   28,   27,   26,   17,  -45,
	 -65,  -67,  -13,  -64,  -74,  -82,  -37,  -75,   47,   45,
	  42,   34,   18,   16,   13,  -35,  -72,  -16,   21,   -1,
	 -43,  -77,  -76,   42,  -34,  -80,  -79,   20,   -1
};
static short fktpgo[] = {
	   0,    0,    0,   56,   37,   29,   29,   29,   29,   29,
	  14,   14,   14,   14,   14,   14,   14,   14,   14,   14,
	  14,   45,   37,   37,   37,   37,   47,   47,   56,   56,
	  49,   49,   49,   50,   50,   52,   54,   54,   37,   45,    0
};
static short fktrlen[] = {
	   0,    0,    0,    1,    5,    2,    3,    3,    2,    2,
	   3,    3,    3,    3,    3,    3,    4,    1,    1,    1,
	   1,    2,    7,    6,    5,    3,    1,    1,    4,    3,
	   2,    3,    2,    1,    2,    3,    1,    3,    2,    1,    2
};
#define FKTS0	56
#define FKTDELTA	42
#define FKTNPACT	57
#define FKTNDEF	13

#define FKTr38	0
#define FKTr39	1
#define FKTr40	2
#define FKTr9	3
#define FKTr19	4
#define FKTr22	5
#define FKTr23	6
#define FKTr24	7
#define FKTr25	8
#define FKTr26	9
#define FKTr31	10
#define FKTr32	11
#define FKTr33	12
#define FKTr34	13
#define FKTr35	14
#define FKTr37	15
#define FKTr36	16
#define FKTr30	17
#define FKTr29	18
#define FKTr28	19
#define FKTr27	20
#define FKTr21	21
#define FKTr18	22
#define FKTr17	23
#define FKTr16	24
#define FKTr15	25
#define FKTr13	26
#define FKTr12	27
#define FKTr11	28
#define FKTr10	29
#define FKTr3	30
#define FKTr2	31
#define FKTr1	32
#define FKTrACCEPT	FKTr38
#define FKTrERROR	FKTr39
#define FKTrLR2	FKTr40
#if FKTDEBUG
char * fktstoken[] = {
	"error",
	"ARRAYVARIABLE",
	"MAIN",
	"DECLARE",
	"WHILE",
	"IF",
	"PRINT",
	"ELSE",
	"EVAL",
	"RETURN",
	"BEG",
	"END",
	"DO",
	"REPEAT",
	"DOUBLEVALUE",
	"INTEGERVALUE",
	"VARIABLE",
	"LONGVARIABLE",
	"OP1",
	"OP2",
	"LOGOP",
	"NOT",
	"OPERATOR",
	"FKT",
	"ACOS",
	"ASIN",
	"ATAN",
	"CEIL",
	"COS",
	"COSH",
	"COT",
	"COTH",
	"EXP",
	"FABS",
	"FACT",
	"FLOOR",
	"GAMMA",
	"LN",
	"LOG",
	"RANDOM",
	"SIN",
	"SINH",
	"SINWT",
	"SQRT",
	"TAN",
	"TANH",
	"RELOP",
	"NE",
	"LS",
	"LE",
	"EQ",
	"GE",
	"GR",
	0
};
char * fktsvar[] = {	"$accept",
	"expression",
	"simplestatement",
	"statement",
	"statementlist",
	"numconstant",
	"program",
	"declarations",
	"declaration",
	"declaration_list",
	"declaration_item",
	0
};
short fktrmap[] = {
	  38,   39,   40,    9,   19,   22,   23,   24,   25,   26,
	  31,   32,   33,   34,   35,   37,   36,   30,   29,   28,
	  27,   21,   18,   17,   16,   15,   13,   12,   11,   10,
	   3,    2,    1,    4,    5,    6,    7,    8,   14,   20,    0
};
short fktsmap[] = {
	   3,   25,   26,   29,   36,   48,   49,   61,   62,   63,
	  64,   65,   75,   79,   76,   71,   68,   67,   66,   60,
	  57,   56,   55,   52,   51,   50,   47,   46,   45,   44,
	  43,   42,   41,   38,   34,   32,   31,   30,   28,   27,
	  20,   19,   17,   15,   14,   13,   12,   11,   10,    9,
	   8,    7,    6,    5,    2,    1,    0,   59,   74,   21,
	  22,   23,   24,   53,   81,   80,   77,   54,   69,   70,
	  78,   72,    4,   40,   18,   16,   39,   58,   37,   73,
	  35,   33
};
int fktntoken = 63;
int fktnvar = 11;
int fktnstate = 82;
int fktnrule = 41;
#endif


// C++ YACC parser code
// Copyright 1991 by Mortice Kern Systems Inc.  All rights reserved.
//
// If FKTDEBUG is defined as 1 and fkt_parse::fktdebug is set to 1,
// fktparse() will print a travelogue of its actions as it reads
// and parses input.
//
// FKTSYNC can be defined to cause fktparse() to attempt to always
// hold a lookahead token

const FKT_MIN_STATE_NUM = 20;	// not useful to be too small!

// Constructor for fkt_parse: user-provided tables
fkt_parse::fkt_parse(int sz, short * states, FKTSTYPE * stack)
{
	mustfree = 0;
	if ((size = sz) < FKT_MIN_STATE_NUM
	 || (stateStack = states) == (short *) 0
	 || (valueStack = stack) == (FKTSTYPE *) 0) {
		fprintf(stderr,"Bad state/stack given");
		exit(1);
	}
	reset = 1;		// force reset
#if FKTDEBUG
	fktdebug = 0;
#endif
}
// Constructor for fkt_parse: allocate tables with new
fkt_parse::fkt_parse(int sz)
{
	size = sz;
	reset = 1;		// force reset
	mustfree = 1;		// delete space in deconstructor
#if FKTDEBUG
	fktdebug = 0;
#endif
	stateStack = new short[size+1];
	valueStack = new FKTSTYPE[size+1];
	if (stateStack == (short *) 0 || valueStack == (FKTSTYPE *) 0) {
		printf("Not enough space for parser stacks");
		exit(1);
	}
}
// Destructor for class fkt_parse
//	Free up space
fkt_parse::~fkt_parse()
{
	if (mustfree) {
		delete stateStack;
		delete valueStack;
	}
	stateStack = (short *) 0;
}

// The parser proper.
//	Note that this code is reentrant; you can return a value
//	and then resume parsing by recalling fktparse().
//	Call fktreset() before fktparse() if you want a fresh start
int
fkt_parse::fktparse(fkt_scan* ps)
{
	short	* fktp, * fktq;		// table lookup
	int	fktj;

	if ((scan = ps) == (fkt_scan *) 0) {	// scanner
		fprintf(stderr,"No scanner");
		exit(1);
	}

	if (reset) {			// start new parse
		fktnerrs = 0;
		fkterrflag = 0;
		fktps = stateStack;
		fktpv = valueStack;
		fktstate = FKTS0;
		fktclearin();
		reset = 0;
	} else			// continue saved parse
		goto fktNext;			// after action

fktStack:
	if (++fktps > &stateStack[size]) {
		scan->fkterror("Parser stack overflow");
		FKTABORT;
	}
	*fktps = fktstate;	/* stack current state */
	*++fktpv = fktval;	/* ... and value */

#if FKTDEBUG
	if (fktdebug)
		fktShowState();
#endif

	/*
	 * Look up next action in action table.
	 */
fktEncore:
#ifdef FKTSYNC
	if (fktchar < 0) {
		if ((fktchar = scan->fktlex()) < 0)
			fktchar = 0;
		fktlval = ::fktlval;
#if FKTDEBUG
		if (fktdebug)
			fktShowRead();	// show new input token
#endif
	}
#endif
	if (fktstate >= sizeof fktpact/sizeof fktpact[0]) 	/* simple state */
		fkti = fktstate - FKTDELTA;	/* reduce in any case */
	else {
		if(*(fktp = &fktact[fktpact[fktstate]]) >= 0) {
			/* Look for a shift on fktchar */
#ifndef FKTSYNC
			if (fktchar < 0) {
				if ((fktchar = scan->fktlex()) < 0)
					fktchar = 0;
				fktlval = ::fktlval;
#if FKTDEBUG
				if (fktdebug)
					fktShowRead();	// show new input token
#endif
			}
#endif
			fktq = fktp;
			fkti = fktchar;
			while (fkti < *fktp++)
				;
			if (fkti == fktp[-1]) {
				fktstate = ~fktq[fktq-fktp];
#if FKTDEBUG
				if (fktdebug)
					fktShowShift();
#endif
				fktval = fktlval;		/* stack value */
				fktclearin();		/* clear token */
				if (fkterrflag)
					fkterrflag--;	/* successful shift */
				goto fktStack;
			}
		}

		/*
	 	 *	Fell through - take default action
	 	 */

		if (fktstate >= sizeof fktdef /sizeof fktdef[0])
			goto fktError;
		if ((fkti = fktdef[fktstate]) < 0)	 { /* default == reduce? */

			/* Search exception table */
			fktp = &fktex[~fkti];
#ifndef FKTSYNC
			if (fktchar < 0) {
				if ((fktchar = scan->fktlex()) < 0)
					fktchar = 0;
				fktlval = ::fktlval;
#if FKTDEBUG
				if (fktdebug)
					fktShowRead();	// show new input token
#endif
			}
#endif
			while((fkti = *fktp) >= 0 && fkti != fktchar)
				fktp += 2;
			fkti = fktp[1];
		}
	}

#ifdef FKTLR2
fktReduce:	/* reduce fkti */
#endif
	fktj = fktrlen[fkti];

#if FKTDEBUG
	if (fktdebug)
		fktShowReduce(fktj);
#endif
	fktps -= fktj;		/* pop stacks */
	fktpvt = fktpv;		/* save top */
	fktpv -= fktj;
	fktval = fktpv[1];	/* default action $ = $1 */
	switch (fkti) {		/* perform semantic action */
		
case FKTr1: {	/* program :  MAIN statement */

			_compiled_function=fktpvt[0].np;

} break;

case FKTr2: {	/* program :  MAIN declarations statement */

			_compiled_function=fktpvt[0].np;

} break;

case FKTr3: {	/* program :  error ';' */

         fkterrok();

} break;

case FKTr9: {	/* declaration_item :  LONGVARIABLE */

			FKT_declare_longvariable(fktpvt[0].strvalue,(VALUETYPE) 0,ps->fktlineno);
			delete fktpvt[0].strvalue;
	
} break;

case FKTr10: {	/* declaration_item :  LONGVARIABLE '=' numconstant */

			FKT_declare_longvariable(fktpvt[-2].strvalue,fktpvt[0].doubleval,ps->fktlineno);
			delete fktpvt[-2].strvalue;
	
} break;

case FKTr11: {	/* declaration_item :  LONGVARIABLE '[' numconstant ']' */

	 		if ((VALUETYPE) ((unsigned int) fktpvt[-1].doubleval) != fktpvt[-1].doubleval)
				FKT_error("FKTPARS: cannot use a float value as array size !");
			FKT_declare_arrayvariable(fktpvt[-3].strvalue,(unsigned int) fktpvt[-1].doubleval,ps->fktlineno);
			delete fktpvt[-3].strvalue;
	
} break;

case FKTr12: {	/* numconstant :  INTEGERVALUE */

	 		fktval.doubleval=(VALUETYPE) fktpvt[0].ivalue;
	
} break;

case FKTr13: {	/* numconstant :  DOUBLEVALUE */

	 		fktval.doubleval=(VALUETYPE) fktpvt[0].doubleval;
	
} break;

case FKTr15: {	/* statement :  '{' statementlist '}' */

			fktval.np = fktpvt[-1].np;

} break;

case FKTr16: {	/* statement :  WHILE '(' expression ')' statement */

			fktval.np = new Node(WHILE, fktpvt[-2].np, fktpvt[0].np);

} break;

case FKTr17: {	/* statement :  DO statement WHILE '(' expression ')' */

         fktval.np = new Node(REPEAT, fktpvt[-1].np, fktpvt[-4].np);

} break;

case FKTr18: {	/* statement :  IF '(' expression ')' statement ELSE statement */

			fktval.np = new Node(IF,fktpvt[-4].np,fktpvt[-2].np,fktpvt[0].np);

} break;

case FKTr19: {	/* statement :  IF '(' expression ')' statement */

         fktval.np = new Node(IF,fktpvt[-2].np,fktpvt[0].np);

} break;

case FKTr21: {	/* statementlist :  statementlist statement */

			fktval.np = new Node(';', fktpvt[-1].np, fktpvt[0].np);

} break;

case FKTr22: {	/* simplestatement :  PRINT expression */

			fktval.np = new Node(PRINT,fktpvt[0].np);

} break;

case FKTr23: {	/* simplestatement :  VARIABLE '=' expression */

			if (toupper(fktpvt[-2].svalue)==toupper(V_SINWT_T))
				warning("Illegal assignment of read only variable (SINWT_T).\nOUTPUT MAY BE UNUSUABLE !!");
         fktval.np = new Node('=', new Node(VARIABLE,WordRep(fktpvt[-2].svalue)), fktpvt[0].np);

} break;

case FKTr24: {	/* simplestatement :  LONGVARIABLE '=' expression */

	 		if (FKT_LongVariables.Search(fktpvt[-2].strvalue,VAR_SIMPLE)==NULL)
			{
				char *msg;
				if ((msg=new char[ERR_MSG_LEN])==NULL)
					fatal("FKTPARS: out of memory!!");
				sprintf(msg,"FKTPARS: variable %s not declared or type mismatch!!",fktpvt[-2].strvalue);
				FKT_error(msg);
				delete msg;
			}
         fktval.np = new Node('=', new Node(LONGVARIABLE,fktpvt[-2].strvalue), fktpvt[0].np);
			delete fktpvt[-2].strvalue;

} break;

case FKTr25: {	/* simplestatement :  EVAL expression */

			fktval.np = new Node(EVAL,fktpvt[0].np);

} break;

case FKTr26: {	/* simplestatement :  RETURN expression */

			fktval.np = new Node(EVAL,fktpvt[0].np);

} break;

case FKTr27: {	/* expression :  INTEGERVALUE */

			fktval.np = new Node(INTEGERVALUE, fktpvt[0].ivalue);

} break;

case FKTr28: {	/* expression :  DOUBLEVALUE */

			fktval.np = new Node(DOUBLEVALUE, fktpvt[0].doubleval);

} break;

case FKTr29: {	/* expression :  VARIABLE */

			fktval.np = new Node(VARIABLE, WordRep(fktpvt[0].svalue));

} break;

case FKTr30: {	/* expression :  LONGVARIABLE */

	 		if (FKT_test_type_variable(fktpvt[0].strvalue,VAR_SIMPLE,ps->fktlineno)!=FALSE)
				fktval.np = new Node(LONGVARIABLE, fktpvt[0].strvalue);
			delete fktpvt[0].strvalue;

} break;

case FKTr31: {	/* expression :  expression OP1 expression */

			binary: fktval.np = new Node(fktpvt[-1].ivalue, fktpvt[-2].np, fktpvt[0].np);

} break;

case FKTr32: {	/* expression :  expression OP2 expression */

			goto binary;

} break;

case FKTr33: {	/* expression :  expression LOGOP expression */

			goto binary;

} break;

case FKTr34: {	/* expression :  expression NOT expression */

			goto binary;

} break;

case FKTr35: {	/* expression :  expression RELOP expression */

			goto binary;

} break;

case FKTr36: {	/* expression :  FKT '(' expression ')' */

			fktval.np = new Node(fktpvt[-3].ivalue,fktpvt[-1].np);

} break;

case FKTr37: {	/* expression :  '(' expression ')' */

			fktval.np = fktpvt[-1].np;

} break;
	case FKTrACCEPT:
		FKTACCEPT;
	case FKTrERROR:
		goto fktError;
#ifdef FKTLR2
	case FKTrLR2:
#ifndef FKTSYNC
		if (fktchar < 0) {
			if ((fktchar = scan->fktlex()) < 0)
				fktchar = 0;
			fktlval = ::fktlval;
#if FKTDEBUG
			if (fktdebug)
				fktShowRead();	// show new input token
#endif
		}
#endif
		fktj = 0;
		while(fktlr2[fktj] >= 0) {
			if (fktlr2[fktj] == fktstate && fktlr2[fktj+1] == fktchar
			  && fktlook(fktstate, ::fkt2lex(), fktlr2[fktj+2]))
				break;
			fktj += 3;
		}
		if(fktlr2[fktj] < 0)
			goto fktError;
		if(fktlr2[fktj+2] < 0) {
			fktstate = ~ fktlr2[fktj+2];
			goto fktStack;
		}
		fkti = fktlr2[fktj+2];
		goto fktReduce;
#endif
	}
fktNext:
	/*
	 *	Look up next state in goto table.
	 */

	fktp = &fktgo[fktpgo[fkti]];
	fktq = fktp++;
	fkti = *fktps;
	while (fkti < *fktp++)		/* busy little loop */
		;
	fktstate = ~(fkti == *--fktp? fktq[fktq-fktp]: *fktq);
#ifdef FKTTRACE
	fktShowGoto();
#endif
	goto fktStack;

fkterrlabel:	;		/* come here from FKTERROR	*/
	fkterrflag = 1;
	if (fkti == FKTrERROR)
		fktps--, fktpv--;
	
fktError:
	switch (fkterrflag) {

	case 0:		/* new error */
		fktnerrs++;
		fkti = fktchar;
		scan->fkterror("Syntax error");
		if (fkti != fktchar) {
			/* user has changed the current token */
			/* try again */
			fkterrflag++;	/* avoid loops */
			goto fktEncore;
		}

	case 1:		/* partially recovered */
	case 2:
		fkterrflag = 3;	/* need 3 valid shifts to recover */
			
		/*
		 *	Pop states, looking for a
		 *	shift on `error'.
		 */

		for ( ; fktps > stateStack; fktps--, fktpv--) {
			if (*fktps >= sizeof fktpact/sizeof fktpact[0])
				continue;
			fktp = &fktact[fktpact[*fktps]];
			fktq = fktp;
			do
				;
			while (FKTERRCODE < *fktp++);
			if (FKTERRCODE == fktp[-1]) {
				fktstate = ~fktq[fktq-fktp];
				goto fktStack;
			}
				
			/* no shift in this state */
#if FKTDEBUG
			if (fktdebug && fktps > stateStack+1)
				fktShowErrRecovery(fktps[0]);
#endif
			/* pop stacks; try again */
		}
		/* no shift on error - abort */
		break;

	case 3:
		/*
		 *	Erroneous token after
		 *	an error - discard it.
		 */

		if (fktchar == 0)  /* but not EOF */
			break;
#if FKTDEBUG
		if (fktdebug)
			fktShowErrDiscard();
#endif
		fktclearin();
		goto fktEncore;	/* try again in same state */
	}
	FKTABORT;

}


#ifdef FKTLR2
// LR2 fktlook() => lookahead function
int
fktlook(int state, int c2, int i)
{
	int c1 = fktchar;
	short * rsp = fktps;
	int j;
	short *p,*q;
	short *sb,*st;

#if FKTDEBUG
	if (fktdebug) {
	printf("LR2 state %d (%d) char %s (%d) lookahead %s (%d)",
			fktsmap[state],state,fktptok(c1),c1,
			fktptok(c2),c2);
		if(i > 0)
			printf("reduce %d (%d)\n", fktrmap[i], i);
		else
			printf("shift %d (%d)\n", fktsmap[i], i);
	}
#endif
	st = sb = rsp + 1;
	if (i >= 0)
		goto reduce;
  shift:
	state = ~i;
	c1 = c2;
	if (c1 < 0)
		return 1;
	c2 = -1;

  stack:
	if (++st >= &stateStack[size+1]) {
		scan->fkterror("Parser Stack Overflow");
		return 0;
	}
	*st = state;
	if (state >= sizeof fktpact/sizeof fktpact[0])
		i = state - FKTDELTA;
	else {
		p = &fktact[fktpact[state]];
		q = p;
		i = c1;
		while(i < *p++)
			;
		if (i == p[-1]) {
			state = ~q[q-p];
			c1 = c2;
			if (c1 < 0)
				return 1;
			c2 = -1;
			goto stack;
		}
		if(state >= sizeof fktdef/sizeof fktdef[0])
			return 0;
		if((i = fktdef[state]) < 0) {
			p = &fktex[~i];
			while((i = *p) >= 0 && i != c1)
				p += 2;
			i = p[1];
		}
	}
  reduce:
  	j = fktrlen[i];
	if (st-sb >= j)
		st -= j;
	else {
		rsp -= j+st-sb;
		st = sb;
	}
	switch(i) {
	case FKTrERROR:
		return 0;
	case FKTrACCEPT:
		return 1;
	case FKTrLR2:
		j = 0;
		while(fktlr2[j] >= 0) {
			if(fktlr2[j] == state && fktlr2[j+1] == c1)
				if((i = fktlr2[j+2]) < 0)
					goto shift;
				else
					goto reduce;
		}
		return 0;
	}
	p = &fktgo[fktpgo[i]];
	q = p++;
	i = st==sb ? *rsp : *st;
	while(i < *p++);
	state = ~(i == *--p? q[q-p]: *q);
	goto stack;
}
#endif
		
#if FKTDEBUG
	
/*
 *	Print a token legibly.
 *	This won't work if you roll your own token numbers
 */
char *
fkt_parse::fktptok(int i)
{
	static char	buf[10];

	if (i >= FKTERRCODE)
		return fktstoken[i-FKTERRCODE];
	if (i < 0)
		return "";
	if (i == 0)
		return "$end";
	if (i < ' ')
		(void) sprintf(buf, "'^%c'", i+'@');
	else
		(void) sprintf(buf, "'%c'", i);
	return buf;
}
/*
 * Show current state of fktparse
 */
void
fkt_parse::fktShowState()
{
	(void) printf("state %d (%d), char %s (%d)\n%d stateStack entries\n",
		fktsmap[fktstate],fktstate,fktptok(fktchar),fktchar,
		fktpv - valueStack);
}
// show results of reduction: fkti is rule number
void
fkt_parse::fktShowReduce(int npop)
{
	(void) printf("Reduce by rule %d (pop#=%d)\n", fktrmap[fkti], npop);
}
// show read token
void
fkt_parse::fktShowRead()
{
	(void) printf("read %s (%d)\n", fktptok(fktchar), fktchar);
}
// show Goto
void
fkt_parse::fktShowGoto()
{
	(void) printf("goto %d (%d)\n", fktsmap[fktstate], fktstate);
}
// show Shift
void
fkt_parse::fktShowShift()
{
	(void) printf("shift %d (%d)\n", fktsmap[fktstate], fktstate);
}
// show error recovery
void
fkt_parse::fktShowErrRecovery(int pop)
{
	(void) printf("Error recovery pops state %d (%d), uncovers %d (%d)\n",
		fktsmap[pop], pop, fktsmap[fktstate], fktstate);
}
// show token discards in error processing
void
fkt_parse::fktShowErrDiscard()
{
	(void) printf("Error recovery discards %s (%d), ",
		fktptok(fktchar), fktchar);
}
#endif	/* FKTDEBUG */
