typedef union {
   VALUETYPE doubleval;
   INT ivalue;
	CHAR svalue;
	Node *np;
	CHAR *strvalue;
} FKTSTYPE;
const ARRAYVARIABLE = 257;
const MAIN = 258;
const DECLARE = 259;
const WHILE = 260;
const IF = 261;
const PRINT = 262;
const ELSE = 263;
const EVAL = 264;
const RETURN = 265;
const BEG = 266;
const END = 267;
const DO = 268;
const REPEAT = 269;
const DOUBLEVALUE = 270;
const INTEGERVALUE = 271;
const VARIABLE = 272;
const LONGVARIABLE = 273;
const OP1 = 274;
const OP2 = 275;
const LOGOP = 276;
const NOT = 277;
const OPERATOR = 278;
const FKT = 279;
const ACOS = 280;
const ASIN = 281;
const ATAN = 282;
const CEIL = 283;
const COS = 284;
const COSH = 285;
const COT = 286;
const COTH = 287;
const EXP = 288;
const FABS = 289;
const FACT = 290;
const FLOOR = 291;
const GAMMA = 292;
const LN = 293;
const LOG = 294;
const RANDOM = 295;
const SIN = 296;
const SINH = 297;
const SINWT = 298;
const SQRT = 299;
const TAN = 300;
const TANH = 301;
const RELOP = 302;
const NE = 303;
const LS = 304;
const LE = 305;
const EQ = 306;
const GE = 307;
const GR = 308;
extern FKTSTYPE fktlval;

// C++ YACC parser header
// Copyright 1991 by Mortice Kern Systems Inc.  All rights reserved.
//
// fkt_parse => class defining a parsing object
//	fkt_parse needs a class fkt_scan, which defines the scanner.
// %prefix or option -p xx determines name of this class; if not used,
// defaults to 'fkt_scan'
//
// constructor fills in the tables for this grammar; give it a size
//    to determine size of state and value stacks. Default is 150 entries.
// destructor discards those state and value stacks
//
// int fkt_parse::fktparse(fkt_scan *) invokes parse; if this returns,
//	it can be recalled to continue parsing at last point.
// void fkt_parse::fktreset() can be called to reset the parse;
//	call fktreset() before fkt_parse::fktparse(fkt_scan *)
#include <stdio.h>		// uses printf(), et cetera
#include <stdlib.h>		// uses exit()

const FKTERRCODE = 256;		// YACC 'error' value

// You can use these macros in your action code
#define FKTERROR		goto fkterrlabel
#define FKTACCEPT	FKTRETURN(0)
#define FKTABORT		FKTRETURN(1)
#define FKTRETURN(val)	return(val)

class fkt_parse {
protected:
	int	mustfree;	// set if tables should be deleted
	int	size;		// size of state and value stacks
	int	reset;		// if set, reset state
	short	fkti;		// table index
	short	fktstate;	// current state

	short	* stateStack;	// states stack
	FKTSTYPE	* valueStack;	// values stack
	short	* fktps;		// top of state stack
	FKTSTYPE * fktpv;		// top of value stack

	FKTSTYPE fktlval;		// saved fktlval
	FKTSTYPE	fktval;		// $
	FKTSTYPE * fktpvt;	// $n
	int	fktchar;		// current token
	int	fkterrflag;	// error flag
	int	fktnerrs;	// error count
#if FKTDEBUG
	virtual char*	fktptok(int);		// printable token string
	virtual void	fktShowRead();		// see newly read token
	virtual void	fktShowState();		// see state, value stacks
	virtual void	fktShowReduce(int);	// see reduction
	virtual void	fktShowGoto();		// see goto
	virtual void	fktShowShift();		// see shift
	virtual void	fktShowErrRecovery(int);	// see error recovery
	virtual void	fktShowErrDiscard();	// see token discard in error
#endif
public:
	fkt_scan* scan;			// pointer to scanner
	int	fktdebug;	// if set, tracing if compiled with FKTDEBUG=1

	fkt_parse(int = 150);	// constructor for this grammar
	fkt_parse(int, short *, FKTSTYPE *);	// another constructor

	~fkt_parse();		// destructor

	int	fktparse(fkt_scan * ps);	// parse with given scanner

	void	fktreset() { reset = 1; } // restore state for next fktparse()

	void	setdebug(int y) { fktdebug = y; }

#ifdef FKTLR2
	virtual int fktlook(int, int, int);	// LR(2) lookahead
#endif

// The following are useful in user actions:

	void	fkterrok() { fkterrflag = 0; }	// clear error
	void	fktclearin() { fktchar = -1; }	// clear input
	int	FKTRECOVERING() { return fkterrflag != 0; }
};
// end of .hpp header
