%{
//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : FKTPARS.Y
//
// Description:
//		Parser of function compiler.
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <ccpptyp.h>
#include <stdlib.h>
#include "defines.h"
#include "ownerr.h"
#include "Node.h"
#include "SimplVar.h"
#include "ArrayVar.h"
#include "HashTab.h"
#include "FKTScan.h"
#include "FKTPars.h"

// global variables declared in FUNCTION.CPP
extern Node *_compiled_function;
extern HashTable FKT_LongVariables;

// prototypes
INT WordRep(CHAR);
void FKT_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno);
void FKT_declare_arrayvariable(char *varid,unsigned int size,unsigned int lineno);
char FKT_test_type_variable(char *name,char type,unsigned int lineno);

%}

%union {
   VALUETYPE doubleval;
   INT ivalue;
	CHAR svalue;
	Node *np;
	CHAR *strvalue;
};

%token				 ARRAYVARIABLE
%token				 MAIN DECLARE
%token <ivalue>	 WHILE IF PRINT ELSE EVAL RETURN
%token <ivalue>	 BEG END DO	REPEAT

%token <doubleval> DOUBLEVALUE
%token <ivalue>    INTEGERVALUE
%token <svalue>    VARIABLE
%token <strvalue>  LONGVARIABLE
%token <ivalue> 	 OP1
%token <ivalue>	 OP2
%token <ivalue>	 LOGOP
%token <ivalue>	 NOT
%token <ivalue>	 OPERATOR
%token <ivalue>	 FKT ACOS ASIN ATAN CEIL COS COSH COT COTH EXP FABS FACT
%token <ivalue>	 FLOOR GAMMA LN LOG RANDOM SIN SINH SINWT SQRT TAN TANH
%token <ivalue>	 RELOP NE LS LE EQ GE GR

%left RELOP
%left OP1
%left OP2
%left LOGOP
%right NOT

%type <np>			expression
%type <np>			simplestatement
%type <np>			statement statementlist
%type <doubleval> numconstant

%%


program:
	  MAIN statement
     {
			_compiled_function=$2;
     }
	| MAIN declarations statement
     {
			_compiled_function=$3;
     }
	| error ';'
	  {
         fkterrok();
     }
	;

declarations:
	  declaration
	| declarations declaration
	;

declaration:
	  DECLARE declaration_list ';'
	;

declaration_list:
	  declaration_item
	| declaration_list ',' declaration_item
	;

declaration_item:
	  /* simple variable */
	  LONGVARIABLE
	  {
			FKT_declare_longvariable($1,(VALUETYPE) 0,ps->fktlineno);
			delete $1;
	  }
	| LONGVARIABLE '=' numconstant
	  {
			FKT_declare_longvariable($1,$3,ps->fktlineno);
			delete $1;
	  }

	  /* array variable */
	| LONGVARIABLE '[' numconstant ']'
	  {
	 		if ((VALUETYPE) ((unsigned int) $3) != $3)
				FKT_error("FKTPARS: cannot use a float value as array size !");
			FKT_declare_arrayvariable($1,(unsigned int) $3,ps->fktlineno);
			delete $1;
	  }
	;

numconstant:
	  INTEGERVALUE
 	  {
	 		$$=(VALUETYPE) $1;
	  }
	| DOUBLEVALUE
 	  {
	 		$$=(VALUETYPE) $1;
	  }
	;

statement:
	  simplestatement ';'
	| '{' statementlist '}'
	  {
			$$ = $2;
     }
	| WHILE '(' expression ')' statement
	  {
			$$ = new Node(WHILE, $3, $5);
     }
	| DO statement WHILE '(' expression ')'
	  {
         $$ = new Node(REPEAT, $5, $2);
     }
	| IF '(' expression ')' statement ELSE statement
	  {
			$$ = new Node(IF,$3,$5,$7);
     }
	| IF '(' expression ')' statement
	  {
         $$ = new Node(IF,$3,$5);
     }
	;

statementlist:
	  statement
	| statementlist statement
     {
			$$ = new Node(';', $1, $2);
     }
	;

simplestatement:
	  PRINT expression	
     {
			$$ = new Node(PRINT,$2);
     }
	| VARIABLE '=' expression
	  {
			if (toupper($1)==toupper(V_SINWT_T))
				warning("Illegal assignment of read only variable (SINWT_T).\nOUTPUT MAY BE UNUSUABLE !!");
         $$ = new Node('=', new Node(VARIABLE,WordRep($1)), $3);
     }
	| LONGVARIABLE '=' expression
	  {
	 		if (FKT_LongVariables.Search($1,VAR_SIMPLE)==NULL)
			{
				char *msg;
				if ((msg=new char[ERR_MSG_LEN])==NULL)
					fatal("FKTPARS: out of memory!!");
				sprintf(msg,"FKTPARS: variable %s not declared or type mismatch!!",$1);
				FKT_error(msg);
				delete msg;
			}
         $$ = new Node('=', new Node(LONGVARIABLE,$1), $3);
			delete $1;
     }
   | EVAL expression
     {
			$$ = new Node(EVAL,$2);
     }

   | RETURN expression
     {
			$$ = new Node(EVAL,$2);
     }
	;

expression:
	  INTEGERVALUE
     {
			$$ = new Node(INTEGERVALUE, $1);
     }
	| DOUBLEVALUE
     {
			$$ = new Node(DOUBLEVALUE, $1);
     }
	| VARIABLE
     {
			$$ = new Node(VARIABLE, WordRep($1));
     }
	| LONGVARIABLE
     {
	 		if (FKT_test_type_variable($1,VAR_SIMPLE,ps->fktlineno)!=FALSE)
				$$ = new Node(LONGVARIABLE, $1);
			delete $1;
     }
	/*| LONGVARIABLE '[' expression ']'*/

	| expression OP1 expression
	  {
			binary: $$ = new Node($2, $1, $3);
     }
	| expression OP2 expression
	  {
			goto binary;
     }
	| expression LOGOP expression
	  {
			goto binary;
     }
	| expression NOT expression
	  {
			goto binary;
     }
	| expression RELOP expression	
     {
			goto binary;
     }

   | FKT '(' expression ')'
     {
			$$ = new Node($1,$3);
     }
	| '(' expression ')'	
     {
			$$ = $2;
     }
	;

%%

INT WordRep(CHAR ch)
{
	return(ch-(islower(ch) ? 'a' : 'A'));
}

void FKT_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno)
{
	SimpleVariable *help;

	if (FKT_LongVariables.Search(varid,VAR_SIMPLE)!=NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"FKTPARS: variable %s already declared !",varid);
		warning(msg,lineno);
		dealloc_msgmem(msg);
	}
	else
	{
		if ((help=new SimpleVariable(varid,initvalue))==NULL)
			fatal("FKTPARS: out of memory!!",lineno);
		FKT_LongVariables.Insert(help);
	}
}

void FKT_declare_arrayvariable(char *varid,unsigned int size,unsigned int lineno)
{
	ArrayVariable *help;

	if (FKT_LongVariables.Search(varid,VAR_ARRAY)!=NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"FKTPARS: variable %s already declared !",varid);
		warning(msg,lineno);
		dealloc_msgmem(msg);
	}
	else
	{
		if ((help=new ArrayVariable(varid,size))==NULL)
			fatal("FKTPARS: out of memory!!",lineno);
		FKT_LongVariables.Insert(help);
	}
}

char FKT_test_type_variable(char *name,char type,unsigned int lineno)
{
	ListElement *help;
	if (FKT_LongVariables.Search(name,type)==NULL)
	{
		char *msg;
		if ((msg=new char[ERR_MSG_LEN])==NULL)
			fatal("FKTPARS: out of memory!!");
		sprintf(msg,"FKTPARS: variable %s not declared or type mismatch!!",name);
		FKT_error(msg,lineno);
		delete msg;
		return(FALSE);
	}
	return(TRUE);
}
