#ifndef fkt_state_t
#define fkt_state_t unsigned char
#endif
#define YYNEWLINE 10

#include <stdio.h>		// uses printf(), et cetera
#include <stdarg.h>		// uses va_list
#include <stdlib.h>		// uses exit()
#include <string.h>		// uses memmove()

class fkt_scan {
protected:
	fkt_state_t * state;		// state buffer
	int	size;			// length of state buffer
	int	mustfree;		// set if space is allocated
	int	fkt_end;			// end of pushback
	int	fkt_start;		// start state
	int	fkt_lastc;		// previous char
#ifndef YY_PRESERVE	// efficient default push-back scheme
	char save;		// saved fkttext[fktleng]
#else			// slower push-back for fkttext mungers
	char *save;		// saved fkttext[]
	char *push;
#endif

public:
	char * fkttext;		// fkttext text buffer
	FILE   *fktin;			// input stream
	FILE   *fktout;			// output stream
	int	fktlineno;		// line number
	int	fktleng;			// fkttext token length

	fkt_scan(int = 100);	// constructor for this scanner
			// default token & pushback size is 100 bytes
	fkt_scan(int, char*, char*, fkt_state_t*);
				// constructor when tables are given

	~fkt_scan();		// destructor

	int	fktlex();		// begin a scan

	virtual int	fktgetc() {	// scanner source of input characters
		return getc(fktin);
	}

	virtual int	fktwrap() { return 1; }	// EOF processing

	virtual void	fkterror(char *,...);	// print error message

	virtual void	output(int c) { putc(c, fktout); }

	virtual void	YY_FATAL(char * msg) {	// print message and stop
		fkterror(msg); exit(1);
	}
	virtual void	ECHO() {		// print matched input
		fputs((const char *) fkttext, fktout);
	}
	int	input();		// user-callable get-input
	int	unput(int c);		// user-callable unput character
	void	setinput(FILE * in) {		// switch input streams
		fktin = in;
	}
	void	setoutput(FILE * out) {	// switch output
		fktout = out;
	}
	void	NLSTATE() { fkt_lastc = YYNEWLINE; }
	void	YY_INIT() {
		fkt_start = 0;
		fktleng = fkt_end = 0;
		fkt_lastc = YYNEWLINE;
	}
	void	YY_USER() {		// set up fkttext for user
#ifndef YY_PRESERVE
		save = fkttext[fktleng];
#else
		size_t n = fkt_end - fktleng;
		push = save+size - n;
		if (n > 0)
			memmove(push, fkttext+fktleng, n);
#endif
		fkttext[fktleng] = 0;
	}
	void YY_SCANNER() {		// set up fkttext for scanner
#ifndef YY_PRESERVE
		fkttext[fktleng] = save;
#else
		size_t n = save+size - push;
		if (n > 0)
			memmove(fkttext+fktleng, push, n);
		fkt_end = fktleng + n;
#endif
	}
	void	fktless(int n) {		// trim input to 'n' bytes
		if (n >= 0 && n <= fkt_end) {
			YY_SCANNER();
			fktleng = n;
			YY_USER();
		}
	}
	void	fktcomment(char *const mat); // skip comment input
	int	fktmapch(int delim, int escape);	// map C escapes
} ;
