%{
//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : FKTSCAN.L
//
// Description:
//		Scanner of function compiler.
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

#include <ccpptyp.h>

#include "Node.h"
#include "FKTScan.h"
#include "FKTPars.h"
FKTSTYPE	fktlval;

#define TRUE  1
#define FALSE 0

// determine size of table
#define SIZE_OF_TABLE(x) sizeof(x)/sizeof(x[0])

typedef struct
{
   char *token_name;
   unsigned int token_num;
   unsigned int token_attr;
} FKT_TOKEN_TYPE;

// FOLLOWING DATA MUST BE SORTED
FKT_TOKEN_TYPE relops[]=
{
   { "!=" , NE },
   { "<"  , LS },
   { "<=" , LE },
   { "==" , EQ },
   { ">=" , GE },
   { ">"  , GR }
};

// FOLLOWING DATA MUST BE SORTED
FKT_TOKEN_TYPE fkt_tokens[]=
{
   { "ACOS" 	,  FKT		, ACOS  	},
   { "ASIN" 	,  FKT		, ASIN  	},
   { "ATAN" 	,  FKT      , ATAN  	},
   { "BEG" 		,  BEG       			},
   { "CEIL" 	,  FKT      , CEIL  	},
   { "COS"  	,  FKT      , COS   	},
   { "COSH" 	,  FKT      , COSH  	},
   { "COT" 		,  FKT      , COT  	},
   { "COTH" 	,  FKT      , COTH  	},
	{ "DECLARE" ,	DECLARE				},
   { "DO" 		,  DO       			},
   { "ELSE" 	,  ELSE       			},
   { "EVAL" 	,  EVAL       			},
   { "EXP"  	,  FKT      , EXP   	},
   { "FABS" 	,  FKT      , FABS  	},
   { "FACT" 	,  FKT      , FACT  	},
   { "FLOOR"	,  FKT      , FLOOR 	},
   { "GAMMA" 	,  FKT      , GAMMA 	},
   { "IF"   	,  IF         			},
   { "LN"  		,  FKT      , LN   	},
   { "LOG"		,  FKT      , LOG 	},
   { "MAIN"		,	MAIN					},
   { "PRINT"	,  PRINT      			},
   { "RANDOM"	,  FKT		, RANDOM },
   { "RETURN"	,  RETURN     			},
   { "SIN"  	,  FKT      , SIN   	},
   { "SINH" 	,  FKT      , SINH  	},
   { "SINWT" 	,  FKT      , SINWT 	},
   { "SQRT" 	,  FKT      , SQRT  	},
   { "TAN"  	,  FKT      , TAN   	},
   { "TANH" 	,  FKT      , TANH  	},
   { "WHILE"	,  WHILE      			}
};

char search_fkt_token(char *,FKT_TOKEN_TYPE *,unsigned int,unsigned int *,unsigned int *);
%}

letter      [A-Za-z]
digit       [0-9]
hexdigit    [0-9A-Fa-f]
relop       [!=<>]
addop       [+-]
mulop			[/*\^]
logop       [!&|]
brackets    [(){}\[\]]
delimiter   [,;]
i           [iI]
p           [pP]
x           [xX]

%%

[ \t\n]+	      ;  /* ignore blanks,tabs,cr's */

"/*"				{
						/* ignore C-style comments */
						char ch1,ch2;
						char loopexit=FALSE;

						ch1=input();
						ch2=input();
						while (loopexit==FALSE)
						{
							if (ch1=='*')
								if (ch2=='/')
									loopexit=TRUE;
							ch1=ch2;
							ch2=input();
						}
					}

{p}{i}         {
                  fktlval.doubleval=3.14159265;
                  return DOUBLEVALUE;
               }

{delimiter}    {
                  return (fktlval.ivalue=*fkttext);
               }

{digit}+"."{digit}+ {
                  fktlval.doubleval=strtod((char *)fkttext,(char **) 0);
                  return DOUBLEVALUE;
               }

{digit}+	      {
		            fktlval.ivalue = (int) strtol(fkttext, (char **) 0,*fkttext=='0' ? 8 : 10);
		            return INTEGERVALUE;
	            }

0{x}{hexdigit}+ {
		            fktlval.ivalue = (int) strtol(fkttext,(char **) 0, 16);
	             	return INTEGERVALUE;
	            }

{relop}+       {
                  unsigned int retval;
                  unsigned int attribute;

                  if (search_fkt_token((char *)fkttext,relops,SIZE_OF_TABLE(relops),&retval,&attribute))
                  {
                     fktlval.ivalue=retval;
                     return (RELOP);
                  }
                  if ((strlen((char*) fkttext)==1) && (*fkttext=='='))
                     return fktlval.ivalue=*fkttext;
               }

{addop}|{mulop} {
                  fktlval.ivalue = *fkttext;
                  return (((fktlval.ivalue=='+') || (fktlval.ivalue=='-')) ? OP1 : OP2);
               }

{brackets}   	{
                  return fktlval.ivalue = *fkttext;
               }

{logop}     	{
                  fktlval.ivalue = *fkttext;
                  if (fktlval.ivalue=='!')
                     return(NOT);
                  return (LOGOP);
               }

{letter}+{digit}* {
                  unsigned int retval;
                  unsigned int attribute;

                  if (search_fkt_token((char *)fkttext,fkt_tokens,SIZE_OF_TABLE(fkt_tokens),&retval,&attribute))
                  {
                     if (retval==FKT)
                        fktlval.ivalue=attribute;
                     return (retval);
                  }
                  else
                  {
                     if (strlen((char *)fkttext)==1)
                     {
								fktlval.svalue=fkttext[0];
		                  return VARIABLE;
                     }
                     else
							{
								if ((fktlval.strvalue=new char[strlen((char *) fkttext)+1])==NULL)
								{
									fprintf(stderr,"FKTSCAN: out of memory!! Aborting program!!\n");
									exit(1);
								}
								strcpy(fktlval.strvalue,(char *) fkttext);
								return LONGVARIABLE;
							}
	
                  }
               }

.		fkterror("Unknown character %c", *fkttext);

%%

typedef int (*cmpfunc)(const VOID *,const VOID *);

int compare(const PSTR toCompare,const FKT_TOKEN_TYPE *withCompare)
{
   return(stricmp(toCompare,withCompare->token_name));
}

char search_fkt_token(PSTR tokstr,FKT_TOKEN_TYPE *token_list,WORD numoftokens,WORD *ret_val,WORD *attribute)
{
   char found=FALSE;
   FKT_TOKEN_TYPE *itemptr;

   itemptr=(FKT_TOKEN_TYPE *) bsearch(tokstr,token_list,numoftokens,sizeof(FKT_TOKEN_TYPE),(cmpfunc) compare);
   if (itemptr!=NULL)
   {
      found=TRUE;
      *ret_val=itemptr->token_num;
      *attribute=itemptr->token_attr;
   }
   return (found);
}

