//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : FUNCTION.CPP
//
// Description:
//		Implemetation of class 'Function'
//
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <math.h>

#include "defines.h"
#include "InFile.h"
#include "Function.h"
#include "ownerr.h"
#include "mathfunc.h"
#include "SimplVar.h"
#include "HashTab.h"

// global variables
HashTable FKT_LongVariables;
Node *_compiled_function;
char *_funcstring;
unsigned int _funcstringcnt;
InFile SrcFile;
BYTE srcmode;


Fkt_New_Scan::Fkt_New_Scan(unsigned int scanner_table_size)
	: fkt_scan(scanner_table_size)
{
}

int Fkt_New_Scan::fktgetc(void)
{
   char ch;

	switch (srcmode)
	{
		case FUNCSTR  : 	if ((ch=_funcstring[_funcstringcnt])!='\0')
						      {
									_funcstringcnt++;
                           return(ch);
                        }
                        else
	                        return(-1);
								break;
		case FUNCFILE : 	if ((ch=SrcFile.Read())!=EOF)
                           return(ch);
                        else
								{
									SrcFile.Close();
	                        return(-1);
								}
								break;
	}
	return (-1);
}

// constructor
Function::Function()
{
   _funcstring=NULL;
   _funcstringcnt=0;
	FKT_LongVariables.CleanUp();
	RootNode=_compiled_function=NULL;
}

// constructor
Function::Function(char *InputStream,BYTE mode)
{
   _funcstring=NULL;
   _funcstringcnt=0;
	FKT_LongVariables.CleanUp();				// remove Variable-List
	RootNode=_compiled_function=NULL;
	srcmode=mode;
	switch(mode)
	{
		case FUNCSTR  	: NewString(InputStream);
							  break;
		case FUNCFILE 	: NewFile(InputStream);
							  break;
		default			: fatal("unknown mode in FUNCTION declaration");
							  break;
	}
}

// destructor
Function::~Function()
{
	FreeAll(RootNode);
}

// new function
void Function::NewString(char *FuncString)
{
	srcmode=FUNCSTR;
	_funcstring=FuncString;
	_funcstringcnt=0;
   FreeAll(RootNode);
	FKT_LongVariables.CleanUp();
	RootNode=_compiled_function=NULL;
	Compile();
}

// newfile function
void Function::NewFile(char *FileName)
{
	srcmode=FUNCFILE;
	if (!SrcFile.Open(FileName))
		fatal("FATAL:cannot open srcfile");
	else
	{
		FreeAll(RootNode);
		FKT_LongVariables.CleanUp();
		RootNode=_compiled_function=NULL;
		Compile();
	}
}

void Function::Compile(void)
{
	if ((Scan=new Fkt_New_Scan(FKT_SCANNER_TABLE_SIZE))==NULL)
		fatal("FUNCTION: out of memory !!");
   Scan->YY_INIT();
   Parse.fktreset();
	Parse.fktparse(Scan);
	delete Scan;
	InitVariables();
	RootNode=_compiled_function;
}

WORD Function::Optimize(void)
{
	return(Do_Optimize(RootNode,0));
}

// get Function(x)
VALUETYPE Function::operator()(VALUETYPE argument)
{
   return(Eval(V_PARAMETER,argument));
}

VALUETYPE Function::Eval(CHAR var,VALUETYPE value)
{
   variables[var-(islower(var) ? 'a' : 'A')].value=value;
   return(Execute(RootNode));
}

VALUETYPE Function::Execute(Node *node)
{
	VALUETYPE value_help;
	SimpleVariable *var_help;

	switch(node->op) {
	case INTEGERVALUE :	return node->left.value;
	case DOUBLEVALUE  :	return node->left.value;
	case VARIABLE		:	return variables[(WORD) node->left.value].value;
	case LONGVARIABLE	:	var_help=(SimpleVariable *) FKT_LongVariables.Search(node->left.strval,VAR_SIMPLE);
								return var_help->GetValue();

   // arithmetic operators
	case '+'     		:	return Execute(node->LEFT) + Execute(node->RIGHT);
	case '-'     		:	return Execute(node->LEFT) - Execute(node->RIGHT);
	case '*'     		:	return Execute(node->LEFT) * Execute(node->RIGHT);
	case '/'     		:	if ((value_help=Execute(node->RIGHT))==0.0)
									fatal("division by zero in divsion");
							   else
									return (Execute(node->LEFT)/value_help);
	case '^'				: return (pow(Execute(node->LEFT),Execute(node->RIGHT)));
   // end of arithmetic operators

   // bit logic operators
	case '!'     		:	return ((VALUETYPE) (! (WORD) Execute(node->LEFT)));
	case '&'     		:	return ((VALUETYPE) ((WORD) Execute(node->LEFT) & (int) Execute(node->RIGHT)));
	case '|'     		:	return ((VALUETYPE) ((WORD) Execute(node->LEFT) | (int) Execute(node->RIGHT)));
	// case '^'     		:	return ((VALUETYPE) ((WORD) Execute(node->LEFT) ^ (int) Execute(node->RIGHT)));
   // end of logic operators

   // relational operators
	case LS      		:	return Execute(node->LEFT) < Execute(node->RIGHT);
	case GR      		:	return Execute(node->LEFT) > Execute(node->RIGHT);
	case GE      		:	return Execute(node->LEFT) >= Execute(node->RIGHT);
	case LE      		:	return Execute(node->LEFT) <= Execute(node->RIGHT);
	case NE      		:	return Execute(node->LEFT) != Execute(node->RIGHT);
	case EQ      		:	return Execute(node->LEFT) == Execute(node->RIGHT);
   // end of relational operators

   // functions
   case ACOS    		: return (acos(Execute(node->LEFT)));
   case ATAN    		: return (atan(Execute(node->LEFT)));
   case CEIL    		: return (ceil(Execute(node->LEFT)));
   case COS     		: return (cos(Execute(node->LEFT)));
   case COSH    		: return (cosh(Execute(node->LEFT)));
	case COT     		: if ((value_help=tan(Execute(node->LEFT)))==0.0)
									fatal("division by zero in COT");
							  else
									return (1/value_help);
	case COTH     		: if ((value_help=tanh(Execute(node->LEFT)))==0.0)
									fatal("division by zero in COTH");
							  else
									return (1/value_help);
	case EXP     		: return (exp(Execute(node->LEFT)));
	case FABS    		: return (fabs(Execute(node->LEFT)));
	case FACT    		: return (factrl(Execute(node->LEFT)));
	case FLOOR   		: return (floor(Execute(node->LEFT)));
	case GAMMA   		: return (gammln(Execute(node->LEFT)));
	case LN     		: return (log(Execute(node->LEFT)));
	case LOG	   		: return (log10(Execute(node->LEFT)));
	case RANDOM			: randomize(); return( random(Execute(node->LEFT)));
	case SIN     		: return (sin(Execute(node->LEFT)));
	case SINH    		: return (sinh(Execute(node->LEFT)));
	case SQRT    		: return (sqrt(Execute(node->LEFT)));
	case TAN     		: return (tan(Execute(node->LEFT)));
	case TANH    		: return (tanh(Execute(node->LEFT)));
	// end of functions

	case PRINT   		: printf("%f\n", (double) Execute(node->LEFT)); return 0.0;
	case EVAL    		: return(Execute(node->LEFT));
	case ';'     		: Execute(node->LEFT); return Execute(node->RIGHT);
	case '='     		: switch (node->LEFT->val_type)
							  {
									case VALUE_NUMERIC:
										variables[(WORD) node->LEFT->left.value].value=Execute(node->RIGHT);
										break;
									case VALUE_STRING:
										var_help=(SimpleVariable *) FKT_LongVariables.Search(node->LEFT->left.strval,VAR_SIMPLE);
										var_help->SetValue(Execute(node->RIGHT));
									   break;
							  }
							  return(0);
	case WHILE:
							  while (Execute(node->LEFT))
									Execute(node->RIGHT);
							  return 0;
	case REPEAT:
							  do
									Execute(node->RIGHT);
							  while (Execute(node->LEFT));
							  return 0;
	case IF:
							  if (Execute(node->LEFT))
									Execute(node->RIGHT);
							  else if (node->third != (Node *) 0)
									Execute(node->third);
							  return 0;


	// special functions
	// sin(omega*t)
	case SINWT    		: return (sin(PI2/variables[V_SINWT_T-(islower(V_SINWT_T) ? 'a' : 'A')].value*Execute(node->LEFT)));
	}
	printf("Error: bad Node type (%u)\n", node->op);
	exit(1);
	return(0);
}

// set specific variable
void Function::SetParameter(char id,VALUETYPE value)
{
	//if (!variables[id-(islower(id) ? 'a' : 'A')].used_as_parameter)
	//	printf("You cannot assign the nonparameter variable %c\n",id);
	//else
		variables[id-(islower(id) ? 'a' : 'A')].value=value;
}

// Initialize variables
void Function::InitVariables(void)
{
	for (unsigned int i=0;i<26;i++)
	{
		variables[i].value=0.0;
		variables[i].used_as_parameter=0;
	}
	FKT_LongVariables.SetInitialValue();
}

// Discard memory used for program tree
void Function::FreeAll(Node *np)
{
	if (np!=NULL)
	{
		switch(np->op)
		{
		case IF:		// Triple
			if (np->third)
				FreeAll(np->third);
			/* FALLTHROUGH */
				// Binary
		case '+':
		case '-':
		case '*':
		case '/':
		case ';':
		case '<':
		case '>':
		case NE:
		case LS:
		case LE:
		case EQ:
		case GE:
		case GR:
		case WHILE:
		case '=':
			FreeAll(np->RIGHT);
		/* FALLTHROUGH */
		case PRINT:		// Unary
		case ACOS:
		case ATAN:
		case CEIL:
		case COS:
		case COSH:
		case COT:
		case COTH:
		case EXP:
		case FABS:
		case FACT:
		case FLOOR:
		case GAMMA:
		case LN:
		case LOG:
		case RANDOM:
		case SIN:
		case SINH:
		case SQRT:
		case TAN:
		case TANH:

		case '!':
		case EVAL:
			FreeAll(np->LEFT);
		break;
		}
		delete(np);		// discard storage for node
	}
}

WORD Function::Do_Optimize(Node *rootnode,WORD no_optimizations)
{
	if (!rootnode ||
	    (rootnode->op==LONGVARIABLE))
		return(no_optimizations);
	no_optimizations=Do_Optimize(rootnode->LEFT,no_optimizations);
	if ((rootnode->LEFT->op==INTEGERVALUE) || (rootnode->LEFT->op==DOUBLEVALUE))
	{
		if ((rootnode->RIGHT->op==INTEGERVALUE) || (rootnode->RIGHT->op==DOUBLEVALUE))
		{
			no_optimizations++;
			rootnode->left.value=Execute(rootnode);
			rootnode->op=DOUBLEVALUE;

			delete rootnode->LEFT;
			rootnode->LEFT=NULL;
			delete rootnode->RIGHT;
			rootnode->RIGHT=NULL;
		}
	}
	no_optimizations=Do_Optimize(rootnode->RIGHT,no_optimizations);
	return(no_optimizations);
}
