//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : FUNCTION.H
//
// Description:
//		Definition of class 'Function'
//
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __FUNCTION

#include "Node.h"
#include "FKTScan.h"
#include "FKTPars.h"

#define FUNCSTR	0x01
#define FUNCFILE	0x02

// sacnner table size - actually 1 KB
#define FKT_SCANNER_TABLE_SIZE	0x0800

class Fkt_New_Scan : public fkt_scan {
public:
	Fkt_New_Scan(unsigned int scanner_table_size);

protected:
   int fktgetc(void);
};

class Function {
	Node *RootNode;
	variable_struct variables[26];
	Fkt_New_Scan *Scan;
	fkt_parse Parse;

public:
   // constructor
   Function();
   // constructor
	// mode==0x01 -> FuncString is a string containing the source
	// mode==0x02 -> FuncString is the filename of the file containing the source
   Function(char *FuncString,BYTE mode=FUNCSTR);
   // destructor
   ~Function();
   // newstring method
   void NewString(char *FuncString);
   // newfile method
   void NewFile(char *FileName);
   // get Function(x)
   VALUETYPE operator()(VALUETYPE argument);
	// set specific variable
	void SetParameter(char id,VALUETYPE value);
	// Optimize constant terms. Returns number of optimizations found.
	WORD Optimize(void);
protected:
	void InitVariables(void);
	void Compile(void);
	VALUETYPE Eval(char var,VALUETYPE value);
	VALUETYPE Execute(Node *);
	void FreeAll(Node *);
	WORD Do_Optimize(Node *,WORD);
};

#define __FUNCTION
#endif
