//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : HASHTAB.CPP
//
// Description:
//		Implementation of class HashTable.
//
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include "HashTab.h"

// constructor
HashTable::HashTable(void)
{
}

// destructor
HashTable::~HashTable(void)
{
	CleanUp();
}

// table operations
// insert new element
void HashTable::Insert(ListElement *newelement)
{
	HashArray[HashFunction(newelement->GetKey())].Insert(newelement);
}

// search element with specific key
ListElement *HashTable::Search(char *key,char type)
{
	return(HashArray[HashFunction(key)].Search(key,type));
}

// Remove complete external chains
void HashTable::CleanUp(void)
{
	RemoveAll();
}

// Reset all Elements to their predefined values
void HashTable::SetInitialValue(void)
{
	unsigned int i;

	for (i=0; i<MAX_HASH; i++)
		HashArray[i].SetInitialValue();
}

///////////////////////////////////////////////////////////////////////////////
// private methods
///////////////////////////////////////////////////////////////////////////////

// This hashfunction algorithm is taken from the book
// ALGORITHMS - Robert Sedgewick (Second Edition)
// page 233
unsigned int HashTable::HashFunction(char const *key)
{
	unsigned int i;
	unsigned int h;

	h=(unsigned int) key[1];
	for (i=2; i<=strlen(key); i++)
		h=((h*32)+(unsigned int) key[i]) % MAX_HASH;
	return(h);
}

void HashTable::RemoveAll(void)
{
	unsigned int i;

	for (i=0; i<MAX_HASH; i++)
		HashArray[i].CleanUp();
}

