//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAV-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : MAIN.CPP
//
// Description:
//		Startup code and main procedure.
//
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////
#include <ctype.h>
#include <alloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <ConfFile.h>

#include "defines.h"
#include "ownerr.h"
#include "wavscan.h"
#include "wavpars.h"

///////////////////////////////////////////////////
// CFG-file name
#define CFG_FILENAME	"WAVCFG.CFG"
///////////////////////////////////////////////////
// size for scanner tables - actually 4 KB
#define WAV_SCANNER_TABLE_SIZE	0x1000
///////////////////////////////////////////////////

///////////////////////////////////////////////////
// Variables holding private variable identifiers.
// See DEFINES.H for description
char V_SINWT_T				='f';
char V_AMPLITUDE			='a';
char V_PARAMETER			='x';
char V_SAMPLEINDEX		='s';
char V_STOP_ON_WARNING	='N';
char V_STOP_ON_ERROR		='N';

FILE *fp;

// RESIZE STACK
extern unsigned _stklen=40960U;

// prototypes
void read_CFG_file(void);
void header(void);

class WAV_New_Scan : public wav_scan
{
public:
	WAV_New_Scan(unsigned int scanner_table_size);
protected:
	int wavgetc(void)	{ return(fgetc(fp)); }
	void waverror(char *fmt,...);
};

WAV_New_Scan::WAV_New_Scan(unsigned int scanner_table_size)
	: wav_scan(scanner_table_size)
{
}

void WAV_New_Scan::waverror(char *fmt,...)
{
	char *msg=alloc_msgmem();
	va_list va;

	va_start(va, fmt);
	(void) vsprintf(msg, fmt, va);
	va_end(va);
	error(msg,wavlineno);
	dealloc_msgmem(msg);
}

void main(int argc, char *argv[])
{
	WAV_New_Scan WAV_Scanner(WAV_SCANNER_TABLE_SIZE);
	wav_parse WAV_Parser;

	header();

	if (argc!=2)
	{
		fprintf(stderr,"USAGE: WAV <program>.\n");
		exit(-1);
	}
	if ((fp=fopen(argv[1],"r"))==NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"Cannot open file %s !!\n",argv[1]);
		fatal(msg);
		dealloc_msgmem(msg);
	}
	//printf("START: Available memory = %lu\n",farcoreleft());
	read_CFG_file();
	WAV_Parser.wavparse(&WAV_Scanner);
	fclose(fp);
	//printf("END: Available memory = %lu\n",farcoreleft());
}

void read_CFG_file(void)
{
	ConfFile WAVCFG(CFG_FILENAME,6);
	
	WAVCFG.Add_Item("VAR_SINWT_T"			,CONF_CH,&V_SINWT_T);
	WAVCFG.Add_Item("VAR_AMPLITUDE"		,CONF_CH,&V_AMPLITUDE);
	WAVCFG.Add_Item("VAR_PARAMETER"		,CONF_CH,&V_PARAMETER);
	WAVCFG.Add_Item("VAR_SAMPLEINDEX"	,CONF_CH,&V_SAMPLEINDEX);
	WAVCFG.Add_Item("STOP_ON_ERROR"		,CONF_CH,&V_STOP_ON_ERROR);
	WAVCFG.Last_Item("STOP_ON_WARNING"	,CONF_CH,&V_STOP_ON_WARNING);

	V_STOP_ON_ERROR=toupper(V_STOP_ON_ERROR);
	V_STOP_ON_WARNING=toupper(V_STOP_ON_WARNING);
}

void header(void)
{
	fprintf(stderr,"\n  P R O J E C T  :  C O C H L E A\n\n");
	fprintf(stderr,"WAVGL - a WAVE-file generation language\n");
	fprintf(stderr,"     Version 0.99b BETA (09.12.93)\n\n");
	fprintf(stderr,"      designed and programmed by\n");
	fprintf(stderr,"       Michael Mller-Wernhart\n");
	fprintf(stderr,"     (C)opyright Vienna 1993-94\n\n\n");
}
