        Page    60,132
;-----------------------------------------------------------------------------
;
; SHUTDOWN.ASM	 Copyright (c) 1995  Robert Collins
;
;       You have my permission to copy and distribute this software for
;       non-commercial purposes.  Any commercial use of this software or
;       source code is allowed, so long as the appropriate copyright
;       attributions (to me) are intact, *AND* my email address is properly
;       displayed.
;
;       Basically, give me credit, where credit is due, and show my email
;       address.
;
;-----------------------------------------------------------------------------
;
;	Robert R. Collins		email:	rcollins@x86.org
;       7201 Avalon Dr.
;       Plano, TX  75025
;
;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; Compiler directives
;-----------------------------------------------------------------------------
	SubTitle   _ShutdownCPUID_
        .radix  16
	.model	small
	.586p


;-----------------------------------------------------------------------------
; External segment definition
;-----------------------------------------------------------------------------
	ABS0		SEGMENT AT 0
	org	4 * 6
	INT06		dd	?

        ABS0            ENDS


        BIOSDATA        SEGMENT AT 40h
                org     67h
        Shut_Restart    label   dword
	BIOSDATA	ENDS

	BIOS		SEGMENT at 0F000h
		org	0fff0h
	ResetVector	label	far
	BIOS	ends


;-----------------------------------------------------------------------------
; Public declarations
;-----------------------------------------------------------------------------
	Public	_ShutdownCPUID_,	_CheckShutdownCPUID_,	OBJ_SHUTDOWN


;-----------------------------------------------------------------------------
; External declarations
;-----------------------------------------------------------------------------
	Extern	Signature	:	Word	; Defined in CPUID.ASM
	Extern	Enable_gate20	:	Near	; Defined in KBC.ASM
	Extern	Shut_a20	:	Near	; Defined in KBC.ASM


;-----------------------------------------------------------------------------
; Include data structures, equates, and local variables
;-----------------------------------------------------------------------------
%	Include INCLUDEDIR\\struc.inc	; Include structures
%	Include INCLUDEDIR\\macros.inc	; Include macros
%	Include INCLUDEDIR\\equates.inc ; Include local equates & I/O ports


	.data
;-----------------------------------------------------------------------------
; Equates & local variables
;-----------------------------------------------------------------------------


;-----------------------------------------------------------------------------
; Text equates
;-----------------------------------------------------------------------------
	DescTbl 	equ	(Descriptor ptr [ebx])


;-----------------------------------------------------------------------------
; Misc. local variables
;-----------------------------------------------------------------------------
; Global descriptor table
;-----------------------------------------------------------------------------
	align
	Global_GDT	label	fword		; SELECTOR = 00h
	GDT_Sel 	equ $-Global_GDT
	Descriptor	<>			; Pointer to GDT

;-----------------------------------------------------------------------------
; The following GDT entry points to this code segment as USE16
;-----------------------------------------------------------------------------
	RM_CS		equ $-Global_GDT	; SELECTOR = 08h
	Descriptor	<>			; 16-bit CS to this ROM

;-----------------------------------------------------------------------------
; The following GDT entries point to my data segment and stack segment
;-----------------------------------------------------------------------------
	RM_DS		equ $-Global_GDT	; SELECTOR = 10h
	Descriptor	<>			; Real-mode compatible DS
	RM_SS		equ $-Global_GDT	; SELECTOR = 18h
	Descriptor	<>			; Real-mode compatible SS

;-----------------------------------------------------------------------------
; The following GDT entries point to a 4 Giga-byte flat data segment
;-----------------------------------------------------------------------------
	DS_4G		equ $-Global_GDT	; SELECTOR = 20h
	Descriptor	<>			; 4G DS
GDT_Len equ	$-Global_GDT


;-----------------------------------------------------------------------------
; Save locations for shutdown values
;-----------------------------------------------------------------------------
	StackPtr	dd	?
	i8259_Amask	db	?
	i8259_Bmask	db	?
	StackBuf	dw	3 dup (?)
	OrigEBP 	dd	?
	OrigINT06	dd	?
	RestartType	db	?


	.code
	ASSUME	CS:_TEXT, ds:_DATA, ES:_DATA, FS:NOTHING, GS:NOTHING
	OBJ_SHUTDOWN	label	word
    ifdef	RRC
	org	00214h
    endif

;-----------------------------------------------------------------------------
  _ShutdownCPUID_	proc	near
;-----------------------------------------------------------------------------
; Input:   None
; Output:  AX = CPUID if NC
;	   DX = CPUID if NC (copy of AX)
;	   Nothing if CY
;-----------------------------------------------------------------------------
; Prepare to shutdown the processor
;  * Save SS:SP
;  * Save the restart location in case the invalid opcode handler doesn't get
;    control
;  * Save contents of memory just below 64k.  This is where the default stack
;    will write after a RESET, so we don't want to pollute anything here.
;  * Install an INT06 handler to trap the CPUID upon reset
;  * Save the interrupt masks
;  * Set CMOS shutdown type
;  * Disable A20
;  * Triple-fault the processor
;-----------------------------------------------------------------------------
	mov	ax,ss			; save stack pointer
	shl	eax,10h
	mov	ax,sp
	mov	StackPtr,eax
	mov	OrigEBP,ebp		; saveit

	mov	ax,seg BIOSDATA 	; get BIOS data segment
	mov	fs,ax			; save it
	mov	ax,cs			; save restart vector at 40:67
	shl	eax,10h 		; segment:offset pointer
	mov	ax,offset Restart1
	mov	fs:Shut_Restart,eax

	mov	bx,seg ABS0		; Install new INT06 handler
	mov	fs,bx
	mov	bx,fs:[0fffah]		; get these contents
	mov	StackBuf[0],bx		; save contents
	mov	bx,fs:[0fffch]		; get these contents
	mov	StackBuf[2],bx		; save contents
	mov	bx,fs:[0fffeh]		; get these contents
	mov	StackBuf[4],bx		; save contents

	mov	ebx,fs:INT06		; get original INT06 handler
	mov	OrigINT06,ebx		; save it
	mov	ax,offset RestartINT06	; get INT06 restart location
	mov	fs:INT06,eax

	in	al,21h			; save PIC masks
	mov	i8259_Amask,al
	in	al,0A1h 		; save PIC masks
	mov	i8259_Bmask,al

	mov	al,CMOS_Shutdown	; set shutdown type
	out	Cmos_index,al
	IO_Delay
	mov	al,User_Shutdown
	out	Cmos_data,al

;-----------------------------------------------------------------------------
; Some chipsets snoop the keyboard controller accesses to trap a RESET
; command to the controller.  This is converted to a FAST-RESET by the
; chipset.  Some chipsets, like the Opti 495, require a HLT instruction
; as part of this sequence.  So don't remove the HLT and convert it into
; something else.
;-----------------------------------------------------------------------------
	call	Shut_a20		; disable A20
	mov	RestartType,06h 	; set restart type sentinal
	sti				; enable interrupts
	mov	cx,100h 		; set loop counter
	mov	al,0feh 		; use KBC to issue RESET
	out	KBC_STAT,al		; which ever is your favorite way to
@@:	hlt				; this will guarantee no race condition
	loop	@B
	jmp	ResetVector		; just in case, so we don't hang

;-----------------------------------------------------------------------------
; Regardless of whether or not we restart from 40:67 or the INT06 handler,
; we need to restart the state of the processor to a working state.
;-----------------------------------------------------------------------------
Restart1:
	mov	bx,seg _DATA		; restore data segments
	mov	ds,bx
	mov	RestartType,'R' 	; set restart type

RestartINT06:
	mov	bx,seg _DATA		; restore data segments
	mov	ds,bx
	mov	es,bx
	lss	sp,StackPtr		; restore stack pointer
	mov	ebp,OrigEBP		; restore EBP

	mov	al,CMOS_Shutdown	; clear shutdown type
	out	Cmos_index,al
	IO_Delay
	mov	al,0
	out	Cmos_data,al

	mov	al,i8259_Amask		; restore PIC masks
	out	21h,al
	mov	al,i8259_Bmask		; restore PIC masks
	out	0a1h,al

	mov	bx,seg ABS0		; restore INT06 handler
	mov	fs,bx
	mov	bx,StackBuf[0]		; Restore stack buffer contents
	mov	fs:[0fffah],bx		;
	mov	bx,StackBuf[2]		;
	mov	fs:[0fffch],bx		;
	mov	bx,StackBuf[4]		;
	mov	fs:[0fffeh],bx		;

	mov	ebx,OrigINT06		; get original INT06 handler
	mov	fs:INT06,ebx		; restore original INT06 handler
	cmp	RestartType,06		; did we trap INT06?
	je	@F			; yes
	stc				; set error flag
	ret

@@:	mov	ax,dx			; make a copy of signature
	clc				; clear error flag
	ret
_ShutdownCPUID_ 	endp


;-----------------------------------------------------------------------------
  _CheckShutdownCPUID_	proc	near	; Check for the probability of the
;					; shutdown method to work.
;-----------------------------------------------------------------------------
; If we gate A20 ON, and look at FFFFFFF0, and FFEFFFF0, we should be able to
; determine the probability of the reset method of working.  Located at
; FFFFFFF0 is the reset vector.  There should be a FAR JMP at this address.
; When A20 is gated ON, the data at address at FFEFFFF0 should be different
; than the data at FFFFFFF0.  If the data's the same, then this method will
; not work.  If the data we read at FFEFFFF0 is all FF's, then this method
; will work.  If the data is not all FF's, this method might work, but isn't
; guaranteed.
;-----------------------------------------------------------------------------
; Input:   None
; Output:  AX = FFFF and NC if shutdown method will work.
;	   AX = 0000 and CY if shutdown method will not work (or might not
;		work).
;-----------------------------------------------------------------------------
	cli				; disable interrupts
	mov	ax,seg _DATA
	mov	ds,ax
	mov	es,ax

;-----------------------------------------------------------------------------
; Check if in protected mode.  If so, don't try this technique.
;-----------------------------------------------------------------------------
	smsw	ax			; save low bits of CR0
	test	al,1			; is protected mode enabled?
	jnz	@FailShutdown		; yes, go split
	call	Enable_gate20
	jc	@FailShutdown
	call	Init_GDT
	lgdt	Global_GDT		; load GDT
	mov	eax,cr0 		; get control register
	or	al,1			; enable protected mode
	mov	cr0,eax 		; now in protected mode
	jmp	@F			; flush prefetch queue
@@:	mov	bx,DS_4G		; make flat (big) real mode
	mov	gs,bx

;-----------------------------------------------------------------------------
; In order for the shutdown method to work, when a CPU RESET occurs, we need
; to fetch garbage, which invokes the invalid opcode handler.	This in turn
; will give us the CPUID in DX.  So to make sure this method will work, we
; need to do a few tests first.
;
; If the BIOS is decoded at FFFFxxxx and FFEFxxxx, then this method won't
; work.  So, let's go about trying to detect this condition.
;-----------------------------------------------------------------------------
; Read from the reset vector at FFFFFFF0, and FFEFFFF0.  If both values
; compare the same, then this method won't work.
;-----------------------------------------------------------------------------
	mov	esi,0fffffff0h
	mov	edi,0ffeffff0h
	mov	edx,gs:[edi]
	mov	ecx,gs:[esi]
	mov	bx,RM_DS
	mov	gs,bx
	mov	eax,cr0 		; get control register
	and	al,not 1		; clear PE
	mov	cr0,eax 		; now back in real mode
	jmp	@F

@@:	cmp	ecx,edx 		; are they the same value?
	jne	@F			; nope, OK so far.

;-----------------------------------------------------------------------------
; Well, we failed the first test.  Go split.
;-----------------------------------------------------------------------------
@FailShutdown:
	xor	ax,ax			; set error condition for high-level
					;  languages
	stc				; set error condition
	ret				; split

;-----------------------------------------------------------------------------
; Just because the two reset vectors didn't equal each other, doesn't mean
; this method will work.  All we did so far, was guarantee that the same
; data isn't decoded twice.  That doesn't mean that another ROM isn't there.
;
; So we want to increase the liklihood that a ROM isn't there.  If the BUS
; is floating, then we'll read garbage.  If it's garbage, then chances are
; pretty good that the data we just read didn't begin with the opcode for
; a jump instruction.  If the bus isn't floating, then we'll read real data.
; If a ROM doesn't exist, we'll probably read FFFF, but if a ROM does exist,
; then we'll read the ROM contents.
;
; So to minimize the possiblity of a ROM being decoded, we'll look for a
; jump opcode at the beginning of the decoded address.	If a valid jmp opcode
; exists, then we'll assume that we can't use this code, and split.
;-----------------------------------------------------------------------------
@@:	cmp	dl,0eah 		; FAR JMP?
	je	@FailShutdown		; yes
	cmp	dl,0e9h 		; NEAR JMP?
	je	@FailShutdown		; yes
	mov	ax,-1			; set result status for high-level
					;  languages
	clc
	ret
_CheckShutdownCPUID_	endp



;-----------------------------------------------------------------------------
  Make_Descript 	proc	near	; Initialize a descriptor table entry
;-----------------------------------------------------------------------------
; Input:   EAX = Base address
;	   EBX = Pointer to GDT base address
;	   ECX = Limit (4G = 0FFFFFFFFh)
;	    DL = Access rights
;	    DH = Default operand size (for code segments only)
;	   EDI = Segment selector to access
;-----------------------------------------------------------------------------
	mov	DescTbl[edi].Base_A15_A00,ax
	shr	eax,10h 		; now get upper bits of physical addr
	mov	DescTbl[edi].Base_A23_A16,al
	mov	DescTbl[edi].Base_A31_A24,ah

;-----------------------------------------------------------------------------
; Set limit
;-----------------------------------------------------------------------------
	push	dx
	mov	eax,ecx 		; get limit
	mov	ecx,1000h		; initialize loop counter
	xor	edx,edx 		; clear register
	cmp	eax,100000h		; is limit > 1M?
	jbe	@F			; yes, continue
	test	eax,0fffh		; prepare to round off limit
	jz	@1			; no need to round limit
	add	eax,ecx 		; increment to next 4k limit
	adc	edx,0			; adjust for carry
@1:	idiv	ecx			; now rotate bits
	or	eax,800000h		; set granularity bit
@@:	dec	eax
	mov	DescTbl[edi].Seg_limit,ax
	shr	eax,10h 		; get high bits of limit
	pop	dx			; retore operand size
	or	al,dh			;
	mov	DescTbl[edi].GDLimit_A19_A16,al
	mov	DescTbl[edi].Access_rights,dl
	ret
Make_Descript	endp


;-----------------------------------------------------------------------------
  Init_GDT	proc	near	; Initialize the GDT
;-----------------------------------------------------------------------------
; Input:   None
; Output:  None
;-----------------------------------------------------------------------------
	mov	ebx,offset Global_GDT	; get GDT pointer

;-----------------------------------------------------------------------------
; Initialize GDT as a pointer to itself.
;-----------------------------------------------------------------------------
	mov	eax,seg _DATA
	shl	eax,4			; make into physical address
	add	eax,offset Global_GDT
	mov	ecx,GDT_Len		; length of GDT
	mov	dx,0			; DS access rights
	mov	edi,GDT_Sel		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize CS descriptor
;-----------------------------------------------------------------------------
	mov	eax,cs			; get code selector
	shl	eax,4			; make physical address
	mov	ecx,10000h		; 64k limit
	mov	dx,Code_Seg_Access	; CS access rights
	mov	edi,RM_CS		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize DS descriptor
;-----------------------------------------------------------------------------
	mov	eax,seg _DATA
	shl	eax,4			; make into physical address
	mov	ecx,10000h		; 64k limit
	mov	dx,Data_Seg_Access	; DS access rights
	mov	edi,RM_DS		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize SS descriptor
;-----------------------------------------------------------------------------
	mov	eax,seg STACK
	shl	eax,4			; make into physical address
	mov	ecx,10000h		; 64k limit
	mov	dx,Data_Seg_Access	; DS access rights
	mov	edi,RM_SS		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize 4G descriptor
;-----------------------------------------------------------------------------
	mov	eax,0
	mov	ecx,0FFFFFFFFh		; 64k limit
	mov	dx,Data_Seg_Access	; DS access rights
	mov	edi,DS_4G		; selector offset
	call	Make_Descript
	ret
Init_GDT	endp

	.stack

	end
