#ifndef WAVDEBUG
#define WAVDEBUG 0
#endif

//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : WAVPARS.Y
//
// Description:
//		Parser of WAV-generation language.
//
//	Version 1.0
//
// created       : 02.10.1993
// last modified : 06.12.1993
//
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// special comments
//
//	'ps->' in function-calls warning,error and fatal references to the scanner.
//        This pointer is automatically created by YACC.
//////////////////////////////////////////////////////////////////////////////
#include <ctype.h>

#include "defines.h"
#include "ownerr.h"
#include "CWAVFile.h"
#include "WavScan.h"
#include "WavPars.h"
#include "SimplVar.h"
#include "HashTab.h"

// verbose modes
#define SWITCH_OFF	0x00
#define SWITCH_ON		0x01

// some error codes

// variables
CWAVFile WAVFile;
HashTable WAV_LongVariables;

char function_defined=FALSE;

// switches
char verbose=SWITCH_OFF;
char autoscale=SWITCH_OFF;

// prototypes
void generate_tone_proc(char channelid,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE startat, VALUETYPE duration);
void WAV_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno);

extern WAVSTYPE wavlval;
void generate_tone_proc(char channelid,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE startat, VALUETYPE duration)
{
	float compression_ratio=DEF_COMPRESSION;

	if (verbose==SWITCH_ON)
	{
		switch(channelid)
		{
			case CHANNEL_MONO:
				printf("Generating tone mono :\n");
				break;
			case CHANNEL_LEFT:
				printf("Generating tone on left channel :\n");
				break;
			case CHANNEL_RIGHT:
				printf("Generating tone on right channel:\n");
				break;
		}
		printf("   frequency         : %7.3f Hz\n",frequency);
		printf("   amplitude         : %7.3f\n",amplitude);
		printf("   start at          : %7.3f sec\n",startat);
		printf("   duration          : %7.3f sec\n",duration);
		printf("\n");
		printf("   interval          :[%7.3f - %7.3f] sec\n",startat,startat+duration);

		if (amplitude > 0.0)
		{
			if (autoscale==SWITCH_ON)
			{
				printf("Calculating amplitude compression ratio...");
				compression_ratio=WAVFile.CalcToneCompression(channelid,frequency,amplitude,startat,duration,compression_ratio);
				if (verbose==SWITCH_ON)
					printf(" --> compression ratio = %7.2f\n",compression_ratio);
			}
			printf("Generating tone...\n");
			WAVFile.GenTone(channelid,frequency,amplitude,startat,duration,compression_ratio);
		}
	}
}

void WAV_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno)
{
	SimpleVariable *help;

	if (WAV_LongVariables.Search(varid,VAR_SIMPLE)!=NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"WAVPARS: variable %s already declared !",varid);
		warning(msg,lineno);
		dealloc_msgmem(msg);
	}
	else
	{
		if ((help=new SimpleVariable(varid,initvalue))==NULL)
			fatal("WAVPARS: out of memory!!",lineno);
		WAV_LongVariables.Insert(help);
	}
}
static short wavdef[] = {
	  -1,   -7,   11,    3,   66,   11,    7,    5,    4,    9,
	   8,    6,   10
};
static short wavex[] = {
	 265,   67,  277,   67,   -1,    1,    0,    0,   -1,    1
};
static short wavact[] = {
	 -73,  290,  -75,  -74,  280,  279, -111,   59, -108,  -70,
	 277,  265,  -68,  284,  -58,  -66,  -59,  -60, -109,  -61,
	 -63,  -62,  -65,  -73,  -64,  -67,  296,  294,  290,  275,
	 270,  268,  266,  265,  264,  262,  259,  257,  -56,  295,
	 -55,   61,  -54,   40,  -98,  295,  -53,   61,  -52,  287,
	 -77,  -78,  -76,  281,  276,  274,  -49, -104,  295,  267,
	  -4,  294, -127,   59,  -58,  -66,  -59,  -60, -110,  -61,
	 -63,  -62,  -65,  -73,  -64,  -67,  296,  294,  290,  275,
	 270,  268,  266,  265,  264,  262,  259,  257,  -89,  -88,
	 -47,  289,  285,  276,  -91,  -92,  -90,  294,  293,  292,
	 -44,  260,  -43,  260,  -42,  284,  -41,   61,  -40,   44,
	 -39,   44,  -38,  263,  -36,  -34,  -35,  -37,   47,   45,
	  43,   42,  -36,  -34,  -35,  -37,  -33,  283,   47,   45,
	  43,   42,  -32,   44,  -31,  267,  -30,  286, -103,  295,
	-107,   59,  -97,   41,  -26,  284,  -25,  284,  -36,  -34,
	 -35,  -37,  -85,  -84,  272,  269,   47,   45,   43,   42,
	 -24,   44,  -36,  -37,   47,   42,  -23,   44, -101,  295,
	 -12,  295,  -22,  293,  -99,  278,  -82,  258,  -19,   44,
	 -83,  -36,  -34,  -35,  -37,   47,   45,   43,   42,   37,
	 -18,   44,  -36,  -34,  -35,  -37,  -79,  -80,  283,  273,
	  47,   45,   43,   42, -106,   59,  -14,   44,   -1
};
static short wavpact[] = {
	   1,   12,   63,  107,  109,   63,  118,  118,  118,  164,
	 164,  175,  185,   97,  207,  205,  198,   97,   97,  191,
	 179,  177,   97,  173,  171,  169,  167,  161,  154,  147,
	 145,   97,  143,   97,   97,   97,   97,  141,   97,   61,
	  97,  139,  137,  135,  133,  127,  113,  111,  105,  103,
	 101,   53,   97,   97,   97,   91,   76,    4,   61,   58,
	  53,   53,   49,   47,   45,   43,   41,   39,   26,   13,
	  10,    7,    4
};
static short wavgo[] = {
	 -20,  -21,  -45,  -50,  -51,   60,   51, -100, -126,  -16,
	  17,   13,  -81,  -13,  -29,  -11,  -10,  -94,  -93,  -29,
	  -9,   -8,  -46,   -7,  -17,   54,   53,   52,   40,   38,
	  36,   35,   34,   33,   31,   22,  -27,  -87,   31,  -86,
	 -95,  -48, -102, -105,  -96,   57,  -28,  -15,   -2,  -71,
	 -69,  -57,  -72, -123,    0,   -6,   -3,   56, -113, -112,
	   5, -114, -115, -116, -117, -118, -119, -120, -121, -122,
	  -5, -125, -124,   39,   -1
};
static short wavpgo[] = {
	   0,    0,    0,   72,   72,   66,   42,   69,   24,   24,
	   0,   59,   44,   44,    4,    4,    4,   12,   12,    9,
	   1,    0,   39,   39,   37,   46,   41,   41,   40,   40,
	  40,   24,   24,   24,   53,   68,   67,   42,   65,   64,
	  64,   63,   63,   61,   50,   50,   48,   48,   48,   49,
	  51,   51,   56,   56,   56,   56,   56,   56,   56,   56,
	  56,   56,   70,   70,   47,   59,   62,   49,    0
};
static short wavrlen[] = {
	   0,    0,    0,    1,    3,    3,    0,    3,    3,    3,
	   1,    0,    1,    1,    1,    1,    1,    1,    1,    2,
	   2,    2,    1,    1,    2,    1,    1,    1,    1,    1,
	   1,    3,    3,    1,    2,    5,    2,    1,   11,    6,
	   7,    4,    2,    2,   11,    6,    2,    3,    4,    2,
	   2,    3,    1,    1,    1,    1,    1,    1,    1,    1,
	   1,    1,    1,    3,    1,    1,    2,    0,    2
};
#define WAVS0	0
#define WAVDELTA	61
#define WAVNPACT	73
#define WAVNDEF	13

#define WAVr66	0
#define WAVr67	1
#define WAVr68	2
#define WAVr24	3
#define WAVr25	4
#define WAVr31	5
#define WAVr33	6
#define WAVr36	7
#define WAVr39	8
#define WAVr40	9
#define WAVr52	10
#define WAVr63	11
#define WAVr65	12
#define WAVr64	13
#define WAVr61	14
#define WAVr60	15
#define WAVr59	16
#define WAVr58	17
#define WAVr57	18
#define WAVr56	19
#define WAVr54	20
#define WAVr53	21
#define WAVr51	22
#define WAVr50	23
#define WAVr49	24
#define WAVr48	25
#define WAVr47	26
#define WAVr46	27
#define WAVr45	28
#define WAVr44	29
#define WAVr43	30
#define WAVr42	31
#define WAVr41	32
#define WAVr38	33
#define WAVr37	34
#define WAVr35	35
#define WAVr34	36
#define WAVr32	37
#define WAVr30	38
#define WAVr29	39
#define WAVr28	40
#define WAVr27	41
#define WAVr26	42
#define WAVr20	43
#define WAVr19	44
#define WAVr18	45
#define WAVr3	46
#define WAVr2	47
#define WAVr1	48
#define WAVrACCEPT	WAVr66
#define WAVrERROR	WAVr67
#define WAVrLR2	WAVr68
#if WAVDEBUG
char * wavstoken[] = {
	"error",
	"AUTOSCALE",
	"BIT",
	"BUFFERSIZE",
	"CHANNEL",
	"CREATE",
	"WAV_DECLARE",
	"DEFAULTS",
	"DEFFUNC",
	"S_END",
	"EXPORT",
	"FROM",
	"GENERATE",
	"HZ",
	"IMPORT",
	"IS",
	"KHZ",
	"MSEC",
	"LEFTCH",
	"MESSAGE",
	"MONO",
	"NEW",
	"NORMALIZED",
	"ON",
	"OFF",
	"RIGHTCH",
	"SAMPLERATE",
	"SEC",
	"S_FILE",
	"STEREO",
	"TO",
	"TONE",
	"TRANSPORT",
	"USING",
	"VERBOSE",
	"WITH",
	"FLOATVALUE",
	"INTVALUE",
	"WAV_LONGVARIABLE",
	"STRING",
	"WAV_VARIABLE",
	0
};
char * wavsvar[] = {	"$accept",
	"amplitude",
	"bits",
	"channelid",
	"duration",
	"duration_multiplier",
	"expression",
	"frequency",
	"frequency_multiplier",
	"int_or_float",
	"mono_stereo",
	"normalized",
	"on_off",
	"samplerate",
	"startat",
	"program",
	"opt_verbose",
	"create_new_file",
	"lines",
	"verbose",
	"line",
	"delimiter",
	"autoscale",
	"declare",
	"definefunc",
	"export_import",
	"generate_tone",
	"longvariable_assignment",
	"message",
	"parameters",
	"variable_assignment",
	"declare_list",
	"declare_item",
	0
};
short wavrmap[] = {
	  66,   67,   68,   24,   25,   31,   33,   36,   39,   40,
	  52,   63,   65,   64,   61,   60,   59,   58,   57,   56,
	  54,   53,   51,   50,   49,   48,   47,   46,   45,   44,
	  43,   42,   41,   38,   37,   35,   34,   32,   30,   29,
	  28,   27,   26,   20,   19,   18,    3,    2,    1,    4,
	   6,    7,    8,    9,   10,   11,   12,   13,   14,   15,
	  16,   17,   22,   23,   55,   62,   21,    5,    0
};
short wavsmap[] = {
	   0,    4,   33,   49,   51,   55,   65,   67,   86,   94,
	  95,  106,  109,  125,  120,  118,  117,  116,  114,  110,
	 108,  107,  104,  103,   99,   98,   97,   90,   89,   84,
	  83,   82,   81,   80,   79,   78,   77,   76,   75,   73,
	  72,   71,   70,   69,   68,   66,   60,   59,   47,   46,
	  45,   41,   40,   38,   37,   36,   35,   22,   21,   20,
	  19,   18,   17,   16,   15,   14,   13,   12,   11,    9,
	   3,    2,    1,    5,    6,   42,   43,   44,  121,  122,
	 123,  113,  115,  100,  101,  102,   88,   57,   58,   61,
	  62,   63,   92,   93,   64,    7,   96,   39,  111,  126,
	 105,  112,   85,   48,   52,  124,   91,   10,   34,   56,
	   8,   54,   74,   32,   31,   30,   29,   28,   27,   26,
	  25,   24,   23,   50,   87,  119,   53
};
int wavntoken = 52;
int wavnvar = 33;
int wavnstate = 127;
int wavnrule = 69;
#endif


// C++ YACC parser code
// Copyright 1991 by Mortice Kern Systems Inc.  All rights reserved.
//
// If WAVDEBUG is defined as 1 and wav_parse::wavdebug is set to 1,
// wavparse() will print a travelogue of its actions as it reads
// and parses input.
//
// WAVSYNC can be defined to cause wavparse() to attempt to always
// hold a lookahead token

const WAV_MIN_STATE_NUM = 20;	// not useful to be too small!

// Constructor for wav_parse: user-provided tables
wav_parse::wav_parse(int sz, short * states, WAVSTYPE * stack)
{
	mustfree = 0;
	if ((size = sz) < WAV_MIN_STATE_NUM
	 || (stateStack = states) == (short *) 0
	 || (valueStack = stack) == (WAVSTYPE *) 0) {
		fprintf(stderr,"Bad state/stack given");
		exit(1);
	}
	reset = 1;		// force reset
#if WAVDEBUG
	wavdebug = 0;
#endif
}
// Constructor for wav_parse: allocate tables with new
wav_parse::wav_parse(int sz)
{
	size = sz;
	reset = 1;		// force reset
	mustfree = 1;		// delete space in deconstructor
#if WAVDEBUG
	wavdebug = 0;
#endif
	stateStack = new short[size+1];
	valueStack = new WAVSTYPE[size+1];
	if (stateStack == (short *) 0 || valueStack == (WAVSTYPE *) 0) {
		printf("Not enough space for parser stacks");
		exit(1);
	}
}
// Destructor for class wav_parse
//	Free up space
wav_parse::~wav_parse()
{
	if (mustfree) {
		delete stateStack;
		delete valueStack;
	}
	stateStack = (short *) 0;
}

// The parser proper.
//	Note that this code is reentrant; you can return a value
//	and then resume parsing by recalling wavparse().
//	Call wavreset() before wavparse() if you want a fresh start
int
wav_parse::wavparse(wav_scan* ps)
{
	short	* wavp, * wavq;		// table lookup
	int	wavj;

	if ((scan = ps) == (wav_scan *) 0) {	// scanner
		fprintf(stderr,"No scanner");
		exit(1);
	}

	if (reset) {			// start new parse
		wavnerrs = 0;
		waverrflag = 0;
		wavps = stateStack;
		wavpv = valueStack;
		wavstate = WAVS0;
		wavclearin();
		reset = 0;
	} else			// continue saved parse
		goto wavNext;			// after action

wavStack:
	if (++wavps > &stateStack[size]) {
		scan->waverror("Parser stack overflow");
		WAVABORT;
	}
	*wavps = wavstate;	/* stack current state */
	*++wavpv = wavval;	/* ... and value */

#if WAVDEBUG
	if (wavdebug)
		wavShowState();
#endif

	/*
	 * Look up next action in action table.
	 */
wavEncore:
#ifdef WAVSYNC
	if (wavchar < 0) {
		if ((wavchar = scan->wavlex()) < 0)
			wavchar = 0;
		wavlval = ::wavlval;
#if WAVDEBUG
		if (wavdebug)
			wavShowRead();	// show new input token
#endif
	}
#endif
	if (wavstate >= sizeof wavpact/sizeof wavpact[0]) 	/* simple state */
		wavi = wavstate - WAVDELTA;	/* reduce in any case */
	else {
		if(*(wavp = &wavact[wavpact[wavstate]]) >= 0) {
			/* Look for a shift on wavchar */
#ifndef WAVSYNC
			if (wavchar < 0) {
				if ((wavchar = scan->wavlex()) < 0)
					wavchar = 0;
				wavlval = ::wavlval;
#if WAVDEBUG
				if (wavdebug)
					wavShowRead();	// show new input token
#endif
			}
#endif
			wavq = wavp;
			wavi = wavchar;
			while (wavi < *wavp++)
				;
			if (wavi == wavp[-1]) {
				wavstate = ~wavq[wavq-wavp];
#if WAVDEBUG
				if (wavdebug)
					wavShowShift();
#endif
				wavval = wavlval;		/* stack value */
				wavclearin();		/* clear token */
				if (waverrflag)
					waverrflag--;	/* successful shift */
				goto wavStack;
			}
		}

		/*
	 	 *	Fell through - take default action
	 	 */

		if (wavstate >= sizeof wavdef /sizeof wavdef[0])
			goto wavError;
		if ((wavi = wavdef[wavstate]) < 0)	 { /* default == reduce? */

			/* Search exception table */
			wavp = &wavex[~wavi];
#ifndef WAVSYNC
			if (wavchar < 0) {
				if ((wavchar = scan->wavlex()) < 0)
					wavchar = 0;
				wavlval = ::wavlval;
#if WAVDEBUG
				if (wavdebug)
					wavShowRead();	// show new input token
#endif
			}
#endif
			while((wavi = *wavp) >= 0 && wavi != wavchar)
				wavp += 2;
			wavi = wavp[1];
		}
	}

#ifdef WAVLR2
wavReduce:	/* reduce wavi */
#endif
	wavj = wavrlen[wavi];

#if WAVDEBUG
	if (wavdebug)
		wavShowReduce(wavj);
#endif
	wavps -= wavj;		/* pop stacks */
	wavpvt = wavpv;		/* save top */
	wavpv -= wavj;
	wavval = wavpv[1];	/* default action $ = $1 */
	switch (wavi) {		/* perform semantic action */
		
case WAVr1: {	/* program :  opt_verbose create_new_file lines S_END */

	 		if (verbose==SWITCH_ON)
				printf("\nTerminated successfully.\n");
	
} break;

case WAVr2: {	/* program :  opt_verbose create_new_file S_END */

	 		warning("WAVPARS: new file created, but there was nothing to do !\n",ps->wavlineno);
	
} break;

case WAVr3: {	/* program :  opt_verbose S_END */

	 		warning("WAVPARS:there was absolutly nothing to do !\n",ps->wavlineno);
	
} break;

case WAVr18: {	/* create_new_file :  NEW S_FILE STRING USING DEFAULTS ';' */

	 		if (verbose==SWITCH_ON)
			{
				printf("Generating new file %s using defaults:\n",wavpvt[-3].svalue);
				printf("   mono(1)/stereo(2) : %7d\n",DEF_MODE);
				printf("   samplerate        : %7u Hz\n",DEF_SAMPLERATE);
				printf("   bits/sample       : %7d bit\n",DEF_BITS_P_SAMPLE);
				printf("   duration          : %7.3f sec\n",DEF_NOOFSECONDS);
			}
	 		WAVFile.NewFile(wavpvt[-3].svalue);
			delete wavpvt[-3].svalue;	// remove string
	
} break;

case WAVr19: {	/* create_new_file :  NEW S_FILE STRING mono_stereo ',' samplerate ',' bits ',' duration ';' */

	
	 		unsigned int bytes_p_sample=bytes_p_sample=(wavpvt[-7].ivalue*wavpvt[-3].ivalue)/8;		

	 		if (verbose==SWITCH_ON)
			{
				printf("Generating new file %s :\n",wavpvt[-8].svalue);
				printf("   mono(1)/stereo(2) : %7d\n",wavpvt[-7].ivalue);
				printf("   samplerate        : %7u Hz\n",wavpvt[-5].ivalue);
				printf("   bits/sample       : %7d bit\n",wavpvt[-3].ivalue);
				printf("   duration          : %7.3f sec\n",wavpvt[-1].fvalue);
			}
	 		WAVFile.NewFile(wavpvt[-8].svalue,wavpvt[-7].ivalue,wavpvt[-5].ivalue,wavpvt[-5].ivalue,bytes_p_sample,wavpvt[-3].ivalue,wavpvt[-1].fvalue);
			delete wavpvt[-8].svalue;	// remove string
	
} break;

case WAVr20: {	/* autoscale :  AUTOSCALE on_off */

			autoscale=wavpvt[0].ivalue;
	
} break;

case WAVr24: {	/* declare_item :  WAV_LONGVARIABLE */

			WAV_declare_longvariable(wavpvt[0].svalue,0.0,ps->wavlineno);
			delete wavpvt[0].svalue;
	
} break;

case WAVr25: {	/* declare_item :  WAV_LONGVARIABLE '=' expression */

			WAV_declare_longvariable(wavpvt[-2].svalue,wavpvt[0].fvalue,ps->wavlineno);
			delete wavpvt[-2].svalue;
	
} break;

case WAVr26: {	/* definefunc :  DEFFUNC STRING */

	 		FKT_errorcnt=0;					// reset error count
		  	function_defined=TRUE;
			WORD nooptimizations;

			if (verbose==SWITCH_ON)
				printf("Compiling function...\n");
	 		WAVFile.SetToneFunction(wavpvt[0].svalue);

			if (verbose==SWITCH_ON)
				printf("Optimizing function...");
			// nooptimizations=WAVFile.OptimizeToneFunction();
			if (verbose==SWITCH_ON)
				//printf(" --> %d optimizations found.\n",nooptimizations);
				printf(" --> not available in this version!\n");
			delete wavpvt[0].svalue; // remove string
	
} break;

case WAVr27: {	/* definefunc :  DEFFUNC FROM S_FILE STRING */

	 		FKT_errorcnt=0;					// reset error count
		  	function_defined=TRUE;
			WORD nooptimizations;

			if (verbose==SWITCH_ON)
				printf("Compiling function...\n");
	 		WAVFile.SetToneFunctionFromFile(wavpvt[0].svalue);

			if (verbose==SWITCH_ON)
				printf("Optimizing function...");
			// nooptimizations=WAVFile.OptimizeToneFunction();
			if (verbose==SWITCH_ON)
				//printf(" --> %d optimizations found.\n",nooptimizations);
				printf(" --> not available in this version!\n");
			delete wavpvt[0].svalue; // remove string
	
} break;

case WAVr28: {	/* export_import :  EXPORT channelid CHANNEL TO S_FILE STRING normalized */

	 		if (verbose==SWITCH_ON)
			{
				printf("Exporting data to file %s ",wavpvt[-1].svalue);
				if (wavpvt[0].ivalue==FILE_NORMALIZED)
					printf("normalized ");
				printf("...\n");
			}

	 		WAVFile.ExportAsInteger(wavpvt[-5].ivalue,wavpvt[-1].svalue,wavpvt[0].ivalue);
			delete wavpvt[-1].svalue; // remove string
	
} break;

case WAVr29: {	/* export_import :  IMPORT channelid CHANNEL FROM S_FILE STRING */

	 		warning("WAVPARS: IMPORT not yet implemented !\n",ps->wavlineno);
			delete wavpvt[0].svalue; // remove string
	
} break;

case WAVr30: {	/* generate_tone :  GENERATE TONE channelid ',' frequency ',' amplitude ',' startat ',' duration */

	 		if (FKT_errorcnt>0)
				error("WAVPARS: tone cannot be generated, because of too many errors in function!");
			else
			{
				if (function_defined==FALSE)
					warning("WAVPARS: no tone-function defined !! Using default function a*sin(wt)",ps->wavlineno);
				generate_tone_proc(wavpvt[-8].ivalue,wavpvt[-6].fvalue,wavpvt[-4].fvalue,wavpvt[-2].fvalue,wavpvt[0].fvalue);
			}
	
} break;

case WAVr31: {	/* longvariable_assignment :  WAV_LONGVARIABLE '=' expression */

	 		SimpleVariable *help;

			if ((help=(SimpleVariable *) WAV_LongVariables.Search(wavpvt[-2].svalue,VAR_SIMPLE))==NULL)
			{
				char *msg=alloc_msgmem();
				sprintf(msg,"WAVPARS: variable %s not declared!!",wavpvt[-2].svalue);
				fatal(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			else
				help->SetValue(wavpvt[0].fvalue);
			delete wavpvt[-2].svalue;

} break;

case WAVr32: {	/* normalized :  NORMALIZED */

	 		wavval.ivalue=FILE_NORMALIZED;
	
} break;

case WAVr33: {	/* normalized :  */

	 		wavval.ivalue=FILE_NOT_NORMALIZED;
	
} break;

case WAVr34: {	/* message :  MESSAGE STRING */

			printf("%s\n",wavpvt[0].svalue);
			delete wavpvt[0].svalue; // remove string
	
} break;

case WAVr35: {	/* parameters :  BUFFERSIZE '(' expression SEC ')' */

			WAVFile.NewMem((float) wavpvt[-2].fvalue);
	
} break;

case WAVr36: {	/* variable_assignment :  WAV_VARIABLE '=' expression */

	      if (function_defined==FALSE)
			{
				char *msg=alloc_msgmem();

				sprintf(msg,"WAVPARS: Assignment of variable '%c' in default function a*sin(wt).",wavpvt[-2].svalue[0]);
				warning(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			if ((toupper(wavpvt[-2].svalue[0])==toupper(V_SINWT_T))   	||
				 (toupper(wavpvt[-2].svalue[0])==toupper(V_AMPLITUDE)) 	||
				 (toupper(wavpvt[-2].svalue[0])==toupper(V_PARAMETER)) 	||
				 (toupper(wavpvt[-2].svalue[0])==toupper(V_SAMPLEINDEX)) )
			{
				char *msg=alloc_msgmem();

				sprintf(msg,"WAVPARS: Ignoring assignment of reserved variable '%c' (See WAVCFG.CFG).",wavpvt[-2].svalue[0]);
				warning(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			else
				WAVFile.SetToneFunctionParameter(wavpvt[-2].svalue,wavpvt[0].fvalue);
	
} break;

case WAVr37: {	/* verbose :  VERBOSE on_off */

	 		verbose=wavpvt[0].ivalue;
	
} break;

case WAVr38: {	/* expression :  int_or_float */

	 		wavval.fvalue=wavpvt[0].fvalue;
	
} break;

case WAVr39: {	/* expression :  expression '+' expression */

	 		wavval.fvalue=wavpvt[-2].fvalue + wavpvt[0].fvalue;
	
} break;

case WAVr40: {	/* expression :  expression '-' expression */

	 		wavval.fvalue=wavpvt[-2].fvalue - wavpvt[0].fvalue;
	
} break;

case WAVr41: {	/* expression :  expression '*' expression */

	 		wavval.fvalue=wavpvt[-2].fvalue * wavpvt[0].fvalue;
	
} break;

case WAVr42: {	/* expression :  expression '/' expression */

	 		if (wavpvt[0].fvalue==0.0)
				fatal("WAVPARS: division by zero !!",ps->wavlineno);
	 		wavval.fvalue=wavpvt[-2].fvalue / wavpvt[0].fvalue;
	
} break;

case WAVr43: {	/* int_or_float :  INTVALUE */

	 		wavval.fvalue=(VALUETYPE) wavpvt[0].ivalue;
	
} break;

case WAVr44: {	/* int_or_float :  FLOATVALUE */

	 		wavval.fvalue=wavpvt[0].fvalue;
	
} break;

case WAVr45: {	/* int_or_float :  WAV_LONGVARIABLE */

	 		SimpleVariable *help;

			if ((help=(SimpleVariable *) WAV_LongVariables.Search(wavpvt[0].svalue,VAR_SIMPLE))==NULL)
			{
				char *msg=alloc_msgmem();
				sprintf(msg,"WAVPARS: variable %s not declared!!",wavpvt[0].svalue);
				fatal(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			else
				wavval.fvalue=help->GetValue();
	
} break;

case WAVr46: {	/* mono_stereo :  MONO */

			wavval.ivalue=MODE_MONO;
	
} break;

case WAVr47: {	/* mono_stereo :  STEREO */

			wavval.ivalue=MODE_STEREO;
	
} break;

case WAVr48: {	/* samplerate :  frequency */

	 		wavval.ivalue=(unsigned int) wavpvt[0].fvalue;
	
} break;

case WAVr49: {	/* frequency :  expression frequency_multiplier */

	 		wavval.fvalue=wavpvt[-1].fvalue*wavpvt[0].fvalue;
	
} break;

case WAVr50: {	/* frequency_multiplier :  HZ */

	 		wavval.fvalue=1.0;
	
} break;

case WAVr51: {	/* frequency_multiplier :  KHZ */

 			wavval.fvalue=1000.0;
	
} break;

case WAVr52: {	/* amplitude :  expression */

	 		if ((wavpvt[0].fvalue < 0.0) || (wavpvt[0].fvalue > 1.0))
			{
				error("WAVPARS: Amplitude must be between 0 and 1\nTone will NOT be generated !!",ps->wavlineno);
				wavpvt[0].fvalue=0.0;
			}
			else
				wavval.fvalue=wavpvt[0].fvalue;
	
} break;

case WAVr53: {	/* amplitude :  expression '%' */

	 		if (wavpvt[-1].fvalue > 100.0)
			{
				error("WAVPARS: Amplitude in % must be between 0 and 100\nTone will NOT be generated !!",ps->wavlineno);
				wavpvt[-1].fvalue=0.0;
			}
			else
				wavval.fvalue=wavpvt[-1].fvalue/100.0;
	
} break;

case WAVr54: {	/* bits :  INTVALUE BIT */

	 		switch (wavpvt[-1].ivalue)
			{
				case 8:
					wavval.ivalue=8;
					break;
				case 12:
				case 16:
					wavval.ivalue=16;
					break;
				default:
					char *msg=alloc_msgmem();

					sprintf(msg,"WAVPARS: Illegal size %i bit\n",wavpvt[-1].ivalue);
					error(msg,ps->wavlineno);
					dealloc_msgmem(msg);
			}
	
} break;

case WAVr56: {	/* duration :  expression duration_multiplier */

	 		wavval.fvalue=wavpvt[-1].fvalue*wavpvt[0].fvalue;
	
} break;

case WAVr57: {	/* duration_multiplier :  SEC */

	 		wavval.fvalue=1.0;
	
} break;

case WAVr58: {	/* duration_multiplier :  MSEC */

	 		wavval.fvalue=0.001;
	
} break;

case WAVr59: {	/* channelid :  MONO */

	 		if (WAVFile.GetMode()!=MODE_MONO)
			{
				warning("WAVPARS: STEREO mode selected! Using LEFT channel!\n",ps->wavlineno);
				wavval.ivalue=CHANNEL_LEFT;
			}
			else
				wavval.ivalue=CHANNEL_MONO;
	
} break;

case WAVr60: {	/* channelid :  LEFTCH */

	 		if (WAVFile.GetMode()!=MODE_STEREO)
			{
				warning("WAVPARS: MONO mode selected! Ignoring channel selection!\n",ps->wavlineno);
				wavval.ivalue=CHANNEL_MONO;
			}
			else
				wavval.ivalue=CHANNEL_LEFT;
	
} break;

case WAVr61: {	/* channelid :  RIGHTCH */

	 		if (WAVFile.GetMode()!=MODE_STEREO)
			{
				warning("WAVPARS: MONO mode selected! Ignoring channel selection!\n",ps->wavlineno);
				wavval.ivalue=CHANNEL_MONO;
			}
			else
				wavval.ivalue=CHANNEL_RIGHT;
	
} break;

case WAVr63: {	/* delimiter :  */

	 		error("WAVPARS: missing semicolon !",ps->wavlineno);
	
} break;

case WAVr64: {	/* on_off :  ON */

			wavval.ivalue=SWITCH_ON;
	
} break;

case WAVr65: {	/* on_off :  OFF */

			wavval.ivalue=SWITCH_OFF;
	
} break;
	case WAVrACCEPT:
		WAVACCEPT;
	case WAVrERROR:
		goto wavError;
#ifdef WAVLR2
	case WAVrLR2:
#ifndef WAVSYNC
		if (wavchar < 0) {
			if ((wavchar = scan->wavlex()) < 0)
				wavchar = 0;
			wavlval = ::wavlval;
#if WAVDEBUG
			if (wavdebug)
				wavShowRead();	// show new input token
#endif
		}
#endif
		wavj = 0;
		while(wavlr2[wavj] >= 0) {
			if (wavlr2[wavj] == wavstate && wavlr2[wavj+1] == wavchar
			  && wavlook(wavstate, ::wav2lex(), wavlr2[wavj+2]))
				break;
			wavj += 3;
		}
		if(wavlr2[wavj] < 0)
			goto wavError;
		if(wavlr2[wavj+2] < 0) {
			wavstate = ~ wavlr2[wavj+2];
			goto wavStack;
		}
		wavi = wavlr2[wavj+2];
		goto wavReduce;
#endif
	}
wavNext:
	/*
	 *	Look up next state in goto table.
	 */

	wavp = &wavgo[wavpgo[wavi]];
	wavq = wavp++;
	wavi = *wavps;
	while (wavi < *wavp++)		/* busy little loop */
		;
	wavstate = ~(wavi == *--wavp? wavq[wavq-wavp]: *wavq);
#ifdef WAVTRACE
	wavShowGoto();
#endif
	goto wavStack;

waverrlabel:	;		/* come here from WAVERROR	*/
	waverrflag = 1;
	if (wavi == WAVrERROR)
		wavps--, wavpv--;
	
wavError:
	switch (waverrflag) {

	case 0:		/* new error */
		wavnerrs++;
		wavi = wavchar;
		scan->waverror("Syntax error");
		if (wavi != wavchar) {
			/* user has changed the current token */
			/* try again */
			waverrflag++;	/* avoid loops */
			goto wavEncore;
		}

	case 1:		/* partially recovered */
	case 2:
		waverrflag = 3;	/* need 3 valid shifts to recover */
			
		/*
		 *	Pop states, looking for a
		 *	shift on `error'.
		 */

		for ( ; wavps > stateStack; wavps--, wavpv--) {
			if (*wavps >= sizeof wavpact/sizeof wavpact[0])
				continue;
			wavp = &wavact[wavpact[*wavps]];
			wavq = wavp;
			do
				;
			while (WAVERRCODE < *wavp++);
			if (WAVERRCODE == wavp[-1]) {
				wavstate = ~wavq[wavq-wavp];
				goto wavStack;
			}
				
			/* no shift in this state */
#if WAVDEBUG
			if (wavdebug && wavps > stateStack+1)
				wavShowErrRecovery(wavps[0]);
#endif
			/* pop stacks; try again */
		}
		/* no shift on error - abort */
		break;

	case 3:
		/*
		 *	Erroneous token after
		 *	an error - discard it.
		 */

		if (wavchar == 0)  /* but not EOF */
			break;
#if WAVDEBUG
		if (wavdebug)
			wavShowErrDiscard();
#endif
		wavclearin();
		goto wavEncore;	/* try again in same state */
	}
	WAVABORT;

}


#ifdef WAVLR2
// LR2 wavlook() => lookahead function
int
wavlook(int state, int c2, int i)
{
	int c1 = wavchar;
	short * rsp = wavps;
	int j;
	short *p,*q;
	short *sb,*st;

#if WAVDEBUG
	if (wavdebug) {
	printf("LR2 state %d (%d) char %s (%d) lookahead %s (%d)",
			wavsmap[state],state,wavptok(c1),c1,
			wavptok(c2),c2);
		if(i > 0)
			printf("reduce %d (%d)\n", wavrmap[i], i);
		else
			printf("shift %d (%d)\n", wavsmap[i], i);
	}
#endif
	st = sb = rsp + 1;
	if (i >= 0)
		goto reduce;
  shift:
	state = ~i;
	c1 = c2;
	if (c1 < 0)
		return 1;
	c2 = -1;

  stack:
	if (++st >= &stateStack[size+1]) {
		scan->waverror("Parser Stack Overflow");
		return 0;
	}
	*st = state;
	if (state >= sizeof wavpact/sizeof wavpact[0])
		i = state - WAVDELTA;
	else {
		p = &wavact[wavpact[state]];
		q = p;
		i = c1;
		while(i < *p++)
			;
		if (i == p[-1]) {
			state = ~q[q-p];
			c1 = c2;
			if (c1 < 0)
				return 1;
			c2 = -1;
			goto stack;
		}
		if(state >= sizeof wavdef/sizeof wavdef[0])
			return 0;
		if((i = wavdef[state]) < 0) {
			p = &wavex[~i];
			while((i = *p) >= 0 && i != c1)
				p += 2;
			i = p[1];
		}
	}
  reduce:
  	j = wavrlen[i];
	if (st-sb >= j)
		st -= j;
	else {
		rsp -= j+st-sb;
		st = sb;
	}
	switch(i) {
	case WAVrERROR:
		return 0;
	case WAVrACCEPT:
		return 1;
	case WAVrLR2:
		j = 0;
		while(wavlr2[j] >= 0) {
			if(wavlr2[j] == state && wavlr2[j+1] == c1)
				if((i = wavlr2[j+2]) < 0)
					goto shift;
				else
					goto reduce;
		}
		return 0;
	}
	p = &wavgo[wavpgo[i]];
	q = p++;
	i = st==sb ? *rsp : *st;
	while(i < *p++);
	state = ~(i == *--p? q[q-p]: *q);
	goto stack;
}
#endif
		
#if WAVDEBUG
	
/*
 *	Print a token legibly.
 *	This won't work if you roll your own token numbers
 */
char *
wav_parse::wavptok(int i)
{
	static char	buf[10];

	if (i >= WAVERRCODE)
		return wavstoken[i-WAVERRCODE];
	if (i < 0)
		return "";
	if (i == 0)
		return "$end";
	if (i < ' ')
		(void) sprintf(buf, "'^%c'", i+'@');
	else
		(void) sprintf(buf, "'%c'", i);
	return buf;
}
/*
 * Show current state of wavparse
 */
void
wav_parse::wavShowState()
{
	(void) printf("state %d (%d), char %s (%d)\n%d stateStack entries\n",
		wavsmap[wavstate],wavstate,wavptok(wavchar),wavchar,
		wavpv - valueStack);
}
// show results of reduction: wavi is rule number
void
wav_parse::wavShowReduce(int npop)
{
	(void) printf("Reduce by rule %d (pop#=%d)\n", wavrmap[wavi], npop);
}
// show read token
void
wav_parse::wavShowRead()
{
	(void) printf("read %s (%d)\n", wavptok(wavchar), wavchar);
}
// show Goto
void
wav_parse::wavShowGoto()
{
	(void) printf("goto %d (%d)\n", wavsmap[wavstate], wavstate);
}
// show Shift
void
wav_parse::wavShowShift()
{
	(void) printf("shift %d (%d)\n", wavsmap[wavstate], wavstate);
}
// show error recovery
void
wav_parse::wavShowErrRecovery(int pop)
{
	(void) printf("Error recovery pops state %d (%d), uncovers %d (%d)\n",
		wavsmap[pop], pop, wavsmap[wavstate], wavstate);
}
// show token discards in error processing
void
wav_parse::wavShowErrDiscard()
{
	(void) printf("Error recovery discards %s (%d), ",
		wavptok(wavchar), wavchar);
}
#endif	/* WAVDEBUG */
