typedef union {
   VALUETYPE fvalue;
   unsigned long ivalue;
	char *svalue;
} WAVSTYPE;
const AUTOSCALE = 257;
const BIT = 258;
const BUFFERSIZE = 259;
const CHANNEL = 260;
const CREATE = 261;
const WAV_DECLARE = 262;
const DEFAULTS = 263;
const DEFFUNC = 264;
const S_END = 265;
const EXPORT = 266;
const FROM = 267;
const GENERATE = 268;
const HZ = 269;
const IMPORT = 270;
const IS = 271;
const KHZ = 272;
const MSEC = 273;
const LEFTCH = 274;
const MESSAGE = 275;
const MONO = 276;
const NEW = 277;
const NORMALIZED = 278;
const ON = 279;
const OFF = 280;
const RIGHTCH = 281;
const SAMPLERATE = 282;
const SEC = 283;
const S_FILE = 284;
const STEREO = 285;
const TO = 286;
const TONE = 287;
const TRANSPORT = 288;
const USING = 289;
const VERBOSE = 290;
const WITH = 291;
const FLOATVALUE = 292;
const INTVALUE = 293;
const WAV_LONGVARIABLE = 294;
const STRING = 295;
const WAV_VARIABLE = 296;
extern WAVSTYPE wavlval;

// C++ YACC parser header
// Copyright 1991 by Mortice Kern Systems Inc.  All rights reserved.
//
// wav_parse => class defining a parsing object
//	wav_parse needs a class wav_scan, which defines the scanner.
// %prefix or option -p xx determines name of this class; if not used,
// defaults to 'wav_scan'
//
// constructor fills in the tables for this grammar; give it a size
//    to determine size of state and value stacks. Default is 150 entries.
// destructor discards those state and value stacks
//
// int wav_parse::wavparse(wav_scan *) invokes parse; if this returns,
//	it can be recalled to continue parsing at last point.
// void wav_parse::wavreset() can be called to reset the parse;
//	call wavreset() before wav_parse::wavparse(wav_scan *)
#include <stdio.h>		// uses printf(), et cetera
#include <stdlib.h>		// uses exit()

const WAVERRCODE = 256;		// YACC 'error' value

// You can use these macros in your action code
#define WAVERROR		goto waverrlabel
#define WAVACCEPT	WAVRETURN(0)
#define WAVABORT		WAVRETURN(1)
#define WAVRETURN(val)	return(val)

class wav_parse {
protected:
	int	mustfree;	// set if tables should be deleted
	int	size;		// size of state and value stacks
	int	reset;		// if set, reset state
	short	wavi;		// table index
	short	wavstate;	// current state

	short	* stateStack;	// states stack
	WAVSTYPE	* valueStack;	// values stack
	short	* wavps;		// top of state stack
	WAVSTYPE * wavpv;		// top of value stack

	WAVSTYPE wavlval;		// saved wavlval
	WAVSTYPE	wavval;		// $
	WAVSTYPE * wavpvt;	// $n
	int	wavchar;		// current token
	int	waverrflag;	// error flag
	int	wavnerrs;	// error count
#if WAVDEBUG
	virtual char*	wavptok(int);		// printable token string
	virtual void	wavShowRead();		// see newly read token
	virtual void	wavShowState();		// see state, value stacks
	virtual void	wavShowReduce(int);	// see reduction
	virtual void	wavShowGoto();		// see goto
	virtual void	wavShowShift();		// see shift
	virtual void	wavShowErrRecovery(int);	// see error recovery
	virtual void	wavShowErrDiscard();	// see token discard in error
#endif
public:
	wav_scan* scan;			// pointer to scanner
	int	wavdebug;	// if set, tracing if compiled with WAVDEBUG=1

	wav_parse(int = 150);	// constructor for this grammar
	wav_parse(int, short *, WAVSTYPE *);	// another constructor

	~wav_parse();		// destructor

	int	wavparse(wav_scan * ps);	// parse with given scanner

	void	wavreset() { reset = 1; } // restore state for next wavparse()

	void	setdebug(int y) { wavdebug = y; }

#ifdef WAVLR2
	virtual int wavlook(int, int, int);	// LR(2) lookahead
#endif

// The following are useful in user actions:

	void	waverrok() { waverrflag = 0; }	// clear error
	void	wavclearin() { wavchar = -1; }	// clear input
	int	WAVRECOVERING() { return waverrflag != 0; }
};
// end of .hpp header
