#ifndef wav_state_t
#define wav_state_t unsigned char
#endif
#define YYNEWLINE 10

#include <stdio.h>		// uses printf(), et cetera
#include <stdarg.h>		// uses va_list
#include <stdlib.h>		// uses exit()
#include <string.h>		// uses memmove()

class wav_scan {
protected:
	wav_state_t * state;		// state buffer
	int	size;			// length of state buffer
	int	mustfree;		// set if space is allocated
	int	wav_end;			// end of pushback
	int	wav_start;		// start state
	int	wav_lastc;		// previous char
#ifndef YY_PRESERVE	// efficient default push-back scheme
	char save;		// saved wavtext[wavleng]
#else			// slower push-back for wavtext mungers
	char *save;		// saved wavtext[]
	char *push;
#endif

public:
	char * wavtext;		// wavtext text buffer
	FILE   *wavin;			// input stream
	FILE   *wavout;			// output stream
	int	wavlineno;		// line number
	int	wavleng;			// wavtext token length

	wav_scan(int = 100);	// constructor for this scanner
			// default token & pushback size is 100 bytes
	wav_scan(int, char*, char*, wav_state_t*);
				// constructor when tables are given

	~wav_scan();		// destructor

	int	wavlex();		// begin a scan

	virtual int	wavgetc() {	// scanner source of input characters
		return getc(wavin);
	}

	virtual int	wavwrap() { return 1; }	// EOF processing

	virtual void	waverror(char *,...);	// print error message

	virtual void	output(int c) { putc(c, wavout); }

	virtual void	YY_FATAL(char * msg) {	// print message and stop
		waverror(msg); exit(1);
	}
	virtual void	ECHO() {		// print matched input
		fputs((const char *) wavtext, wavout);
	}
	int	input();		// user-callable get-input
	int	unput(int c);		// user-callable unput character
	void	setinput(FILE * in) {		// switch input streams
		wavin = in;
	}
	void	setoutput(FILE * out) {	// switch output
		wavout = out;
	}
	void	NLSTATE() { wav_lastc = YYNEWLINE; }
	void	YY_INIT() {
		wav_start = 0;
		wavleng = wav_end = 0;
		wav_lastc = YYNEWLINE;
	}
	void	YY_USER() {		// set up wavtext for user
#ifndef YY_PRESERVE
		save = wavtext[wavleng];
#else
		size_t n = wav_end - wavleng;
		push = save+size - n;
		if (n > 0)
			memmove(push, wavtext+wavleng, n);
#endif
		wavtext[wavleng] = 0;
	}
	void YY_SCANNER() {		// set up wavtext for scanner
#ifndef YY_PRESERVE
		wavtext[wavleng] = save;
#else
		size_t n = save+size - push;
		if (n > 0)
			memmove(wavtext+wavleng, push, n);
		wav_end = wavleng + n;
#endif
	}
	void	wavless(int n) {		// trim input to 'n' bytes
		if (n >= 0 && n <= wav_end) {
			YY_SCANNER();
			wavleng = n;
			YY_USER();
		}
	}
	void	wavcomment(char *const mat); // skip comment input
	int	wavmapch(int delim, int escape);	// map C escapes
} ;
