%{
//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : WAVSCAN.L
//
// Description:
//		Scanner of WAV-generation language.
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

#include <ccpptyp.h>
#include "defines.h"
#include "ownerr.h"
#include "wavscan.h"
#include "wavpars.h"
WAVSTYPE	wavlval;

#define MAXVARNAMESIZE	32
#define MAXSTRINGSIZE	32768

#define TRUE  1
#define FALSE 0

// determine size of table
#define SIZE_OF_TABLE(x) sizeof(x)/sizeof(x[0])

typedef struct
{
   char *token_name;
   unsigned int token_num;
} WAV_TOKEN_TYPE;

// FOLLOWING DATA MUST BE SORTED
WAV_TOKEN_TYPE wav_tokens[]=
{
	{ "AUTOSCALE"	, AUTOSCALE		},
	{ "BIT"			, BIT				},
	{ "BUFFERSIZE"	, BUFFERSIZE	},
	{ "CHANNEL"		, CHANNEL		},
	{ "CREATE"		, CREATE			},
	{ "DECLARE"		, WAV_DECLARE  },
	{ "DEFAULTS"	, DEFAULTS		},
	{ "DEFFUNC"		, DEFFUNC		},
	{ "END"			, S_END			},
	{ "EXPORT"		, EXPORT			},
	{ "FILE"			, S_FILE			},
	{ "FROM"			, FROM 			},
	{ "GENERATE"	, GENERATE		},
	{ "HZ"			, HZ				},
	{ "IMPORT"		, IMPORT			},
	{ "IS"			, IS 				},
	{ "KHZ"			, KHZ				},
	{ "LEFT"			, LEFTCH			},
	{ "MESSAGE"		, MESSAGE		},
	{ "MONO"			, MONO			},
	{ "MSEC"			, MSEC			},
	{ "NEW"			, NEW				},
	{ "NORMALIZED"	, NORMALIZED	},
	{ "OFF"			, OFF				},
	{ "ON"			, ON				},
	{ "RIGHT"		, RIGHTCH		},
	{ "SAMPLERATE" , SAMPLERATE	},
	{ "SEC"			, SEC				},
	{ "STEREO"		, STEREO			},
	{ "TO"			, TO				},
	{ "TONE"			, TONE			},
	{ "USING"		, USING			},
	{ "VERBOSE"		, VERBOSE		},
	{ "WITH"			, WITH			},
};

char search_wav_token(char *,WAV_TOKEN_TYPE *,unsigned int,unsigned int *);
%}
blank		   [ \t\n]
letter      [A-Za-z]
digit       [0-9]
hexdigit    [0-9A-Fa-f]
addop       [+-]
mulop			[/*]
logop       [!&|]
brackets    [(){}]
percent		[%]
delimiter   [,;=]
x				[xX]
other			.

%%

{blank}+	      ;  /* ignore blanks,tabs,cr's */

"-->"				{
						return(TRANSPORT);
					}

"/*"				{
						/* ignore C-style comments */
						char ch1,ch2;
						char loopexit=FALSE;

						ch1=input();
						ch2=input();
						while (loopexit==FALSE)
						{
							if (ch1=='*')
								if (ch2=='/')
									loopexit=TRUE;
							ch1=ch2;
							ch2=input();
						}
					}

{delimiter}|{brackets}|{percent} {
                  return (*wavtext);
               }

"\""				{
						char *str;
						char ch;
						WORD i=0;

						if ((wavlval.svalue=new char[MAXSTRINGSIZE])==NULL)
							fatal("out of memory in STRING constant !!");
						while ((ch=input())!='"')
							wavlval.svalue[i++]=ch;
						wavlval.svalue[i]='\0';

						return STRING;
					}


{digit}+"."{digit}+ {
                  wavlval.fvalue=strtod((char *)wavtext,(char **) 0);
                  return FLOATVALUE;
               }

{digit}+	      {
		            wavlval.ivalue = (unsigned long) strtol(wavtext, (char **) 0,*wavtext=='0' ? 8 : 10);
		            return INTVALUE;
	            }

0{x}{hexdigit}+ {
		            wavlval.ivalue = (unsigned long) strtol(wavtext,(char **) 0, 16);
	             	return INTVALUE;
	            }

{brackets}   	{
                  return(wavlval.ivalue = *wavtext);
               }
{addop}|{mulop} {
						return(wavtext[0]);
					 }

{letter}+{digit}* {
                  unsigned int retval;

                  if (search_wav_token((char *) wavtext,wav_tokens,SIZE_OF_TABLE(wav_tokens),&retval))
                     return (retval);
                  else
						{
							if ((wavlval.svalue=new char[strlen((char *) wavtext)+1])==NULL)
								fatal("out of memory in VARIABLE name !!");
							strcpy(wavlval.svalue,(char *) wavtext);
							return(WAV_LONGVARIABLE);
						}
               }

"F."{letter}+{digit}* {
						int i;

						if ((wavlval.svalue=new char[MAXVARNAMESIZE])==NULL)
							fatal("out of memory in STRING constant !!\n");
						for (i=2; i<strlen((char *) wavtext);i++)
							wavlval.svalue[i-2]=(char) wavtext[i];
						wavlval.svalue[i-2]='\0';

						return WAV_VARIABLE;
					}

{other}			{
						char *msg=alloc_msgmem();
						
						sprintf(msg,"Unknown character : %c", *wavtext);
						error(msg);
						dealloc_msgmem(msg);
					}

%%

typedef int (*cmpfunc)(const void *,const void *);

int compare(const char *toCompare,const WAV_TOKEN_TYPE *withCompare)
{
   return(stricmp(toCompare,withCompare->token_name));
}

char search_wav_token(char *tokstr,WAV_TOKEN_TYPE *token_list,unsigned int numoftokens,unsigned int *ret_val)
{
   char found=FALSE;
   WAV_TOKEN_TYPE *itemptr;

   itemptr=(WAV_TOKEN_TYPE *) bsearch(tokstr,token_list,numoftokens,sizeof(WAV_TOKEN_TYPE),(cmpfunc) compare);
   if (itemptr!=NULL)
   {
      found=TRUE;
      *ret_val=itemptr->token_num;
   }
   return (found);
}
