        Page    60,132
;-----------------------------------------------------------------------------
;
; CPUID.ASM   Copyright (c) 1995  Robert Collins
;
;       You have my permission to copy and distribute this software for
;       non-commercial purposes.  Any commercial use of this software or
;       source code is allowed, so long as the appropriate copyright
;       attributions (to me) are intact, *AND* my email address is properly
;       displayed.
;
;       Basically, give me credit, where credit is due, and show my email
;       address.
;
;-----------------------------------------------------------------------------
;
;	Robert R. Collins		email:	rcollins@x86.org
;       7201 Avalon Dr.
;       Plano, TX  75025
;
;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; Compiler directives
;-----------------------------------------------------------------------------
	Title	_CPUID_
        .radix  16
	.model	small
	.586p


;-----------------------------------------------------------------------------
; External segment definition
;-----------------------------------------------------------------------------
        ABS0            SEGMENT AT 0
        ABS0            ENDS


        BIOSDATA        SEGMENT AT 40h
                org     67h
        Shut_Restart    label   dword
        BIOSDATA        ENDS


;-----------------------------------------------------------------------------
; Include data structures, equates, and local variables
;-----------------------------------------------------------------------------
%	Include INCLUDEDIR\\struc.inc	; Include structures
%	Include INCLUDEDIR\\macros.inc	; Include macros
%	Include INCLUDEDIR\\equates.inc ; Include local equates & I/O ports


	.data
;-----------------------------------------------------------------------------
; Equates & local variables
;-----------------------------------------------------------------------------
; Protected mode access rights
;-----------------------------------------------------------------------------
	CS_access	equ	10011011y
	DS_access	equ	10010011y
	PTE_Present	equ	00000001y	; Page Table
	PTE_RW		equ	00000010y
	PTE_US		equ	00000100y


;-----------------------------------------------------------------------------
; Text equates
;-----------------------------------------------------------------------------
DescTbl 	equ	(Descriptor ptr [ebx])
        CRLF            equ     <0dh,0ah>
        CRLF$           equ     <CRLF,'$'>
        INT6            equ     [bp-4]


;-----------------------------------------------------------------------------
; Misc. local variables
;-----------------------------------------------------------------------------
; Global descriptor table
;-----------------------------------------------------------------------------
	align
	Global_GDT	label	fword		; SELECTOR = 00h
	GDT_Sel 	equ $-Global_GDT
	Descriptor	<>			; Pointer to GDT

;-----------------------------------------------------------------------------
; The following GDT entry points to this code segment as USE16
;-----------------------------------------------------------------------------
	RM_CS		equ $-Global_GDT	; SELECTOR = 08h
	Descriptor	<>			; 16-bit CS to this ROM

;-----------------------------------------------------------------------------
; The following GDT entries point to my data segment and stack segment
;-----------------------------------------------------------------------------
	RM_DS		equ $-Global_GDT	; SELECTOR = 10h
	Descriptor	<>			; Real-mode compatible DS
	RM_SS		equ $-Global_GDT	; SELECTOR = 18h
	Descriptor	<>			; Real-mode compatible SS

;-----------------------------------------------------------------------------
; The following GDT entries point to a 4 Giga-byte flat data segment
;-----------------------------------------------------------------------------
	DS_4G		equ $-Global_GDT	; SELECTOR = 20h
	Descriptor	<>			; 4G DS
GDT_Len equ	$-Global_GDT

        RM_IDT3_Ptr     dw      (256d*4)-1      ; Real-mode IDT
                        dd      0               ;  pointer
                        dw      0

        Result_Status   dw      0
        PIC0_Mask       db      0
        PIC1_Mask       db      0


;-----------------------------------------------------------------------------
; String Messages
;-----------------------------------------------------------------------------


	.code
        ASSUME  CS:_TEXT, ds:_DATA, ES:_DATA, SS:STACK
        ASSUME  FS:NOTHING, GS:NOTHING
;-----------------------------------------------------------------------------
  _CPUID_	proc	far
;-----------------------------------------------------------------------------
	PUSHF
        PUSH    DS                      ; Setup the stack to
        XOR     AX,AX                   ;  return to DOS
        PUSH    AX

;-----------------------------------------------------------------------------
; Code begins here
;-----------------------------------------------------------------------------
	cli				; disable interrupts
	mov	ax,seg _DATA
	mov	ds,ax
	mov	es,ax
	call	Init_GDT
	lgdt	Global_GDT		; load GDT
	mov	eax,cr0 		; get control register
	or	al,1			; enable protected mode
	mov	cr0,eax 		; now in protected mode
	jmp	@F			; flush prefetch queue
@@:	mov	bx,DS_4G		; make flat (big) real mode
	mov	gs,bx
	and	al,not 1		; clear PE
	mov	cr0,eax 		; now back in real mode
	jmp	@F
@@:	call	Enable_gate20
	mov	esi,0fffffff0h
	mov	edi,0ffeffff0h
	mov	ecx,gs:[esi]
	mov	edx,gs:[edi]
	int 3

	lgdt	Global_GDT		; load GDT
	mov	eax,cr0 		; get control register
	or	al,1			; enable protected mode
	mov	cr0,eax 		; now in protected mode
	jmp	@F			; flush prefetch queue
@@:	mov	bx,DS_4G		; make flat (big) real mode
	mov	gs,bx
	and	al,not 1		; clear PE
	mov	cr0,eax 		; now back in real mode
	jmp	@F
@@:	call	Shut_a20
	mov	esi,0fffffff0h
	mov	edi,0ffeffff0h
	mov	ecx,gs:[esi]
	mov	edx,gs:[edi]
	int 3


	iret
_CPUID_ 	endp


;-----------------------------------------------------------------------------
  Make_Descript 	proc	near	; Initialize a descriptor table entry
;-----------------------------------------------------------------------------
; Input:   EAX = Base address
;	   EBX = Pointer to GDT base address
;	   ECX = Limit (4G = 0FFFFFFFFh)
;	    DL = Access rights
;	    DH = Default operand size (for code segments only)
;	   EDI = Segment selector to access
;-----------------------------------------------------------------------------
	mov	DescTbl[edi].Base_A15_A00,ax
	shr	eax,10h 		; now get upper bits of physical addr
	mov	DescTbl[edi].Base_A23_A16,al
	mov	DescTbl[edi].Base_A31_A24,ah

;-----------------------------------------------------------------------------
; Set limit
;-----------------------------------------------------------------------------
	push	dx
	mov	eax,ecx 		; get limit
	mov	ecx,1000h		; initialize loop counter
	xor	edx,edx 		; clear register
	cmp	eax,100000h		; is limit > 1M?
	jbe	@F			; yes, continue
	test	eax,0fffh		; prepare to round off limit
	jz	@1			; no need to round limit
	add	eax,ecx 		; increment to next 4k limit
	adc	edx,0			; adjust for carry
@1:	idiv	ecx			; now rotate bits
	or	eax,800000h		; set granularity bit
@@:	dec	eax
	mov	DescTbl[edi].Seg_limit,ax
	shr	eax,10h 		; get high bits of limit
	pop	dx			; retore operand size
	or	al,dh			;
	mov	DescTbl[edi].GDLimit_A19_A16,al
	mov	DescTbl[edi].Access_rights,dl
	ret
Make_Descript	endp


;-----------------------------------------------------------------------------
  Init_GDT	proc	near	; Initialize the GDT
;-----------------------------------------------------------------------------
; Input:   None
; Output:  None
;-----------------------------------------------------------------------------
	mov	ebx,offset Global_GDT	; get GDT pointer

;-----------------------------------------------------------------------------
; Initialize GDT as a pointer to itself.
;-----------------------------------------------------------------------------
	mov	eax,seg _DATA
	shl	eax,4			; make into physical address
	add	eax,offset Global_GDT
	mov	ecx,GDT_Len		; length of GDT
	mov	dx,0			; DS access rights
	mov	edi,GDT_Sel		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize CS descriptor
;-----------------------------------------------------------------------------
	mov	eax,cs			; get code selector
	shl	eax,4			; make physical address
	mov	ecx,10000h		; 64k limit
	mov	dx,CS_ACCESS		; DS access rights
	mov	edi,RM_CS		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize DS descriptor
;-----------------------------------------------------------------------------
	mov	eax,seg _DATA
	shl	eax,4			; make into physical address
	mov	ecx,10000h		; 64k limit
	mov	dx,DS_ACCESS		; DS access rights
	mov	edi,RM_DS		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize SS descriptor
;-----------------------------------------------------------------------------
	mov	eax,seg STACK
	shl	eax,4			; make into physical address
	mov	ecx,10000h		; 64k limit
	mov	dx,DS_ACCESS		; DS access rights
	mov	edi,RM_SS		; selector offset
	call	Make_Descript

;-----------------------------------------------------------------------------
; Initialize 4G descriptor
;-----------------------------------------------------------------------------
	mov	eax,0
	mov	ecx,0FFFFFFFFh		; 64k limit
	mov	dx,DS_ACCESS		; DS access rights
	mov	edi,DS_4G		; selector offset
	call	Make_Descript
	ret
Init_GDT	endp


	Include 	..\include\KBC.INC

_text	ends


	.stack	800h

	end	_CPUID_
